///////////////////////////////////////////////////////////////////////////////
//    Charles River Analytics Inc.
//    Copyright 2002-2003. All Rights Reserved.
//
//    ORIGINAL SYSTEM: The Vision Kit Library
//    SUBSYSTEM: Image
//    $Workfile: Convolution.h $
//    $Revision: 2 $
//    Author(s): V.Ablavsky, J.Pollak, M.Stevens
//
//    Description
//    ===========
//
// Implements the naive, 4-for median filtering algorithm
//
///////////////////////////////////////////////////////////////////////////////

#include "Image.h"
#include "ImageDefs.h"
#include <vector>
#include <numeric>

#ifndef Convolution_h
#define Convolution_h

namespace cuj
{

  // note sub window dimensions must be odd
  template<size_t _SubWidth, size_t _SubHeight, typename _Tp, typename _KernelType>
    struct Convolution1
  {
    Convolution1(_KernelType *kernel) : kernel_(kernel)
    {
    }
    inline void operator()(const Image<_Tp> &in, Image<_Tp> &out)
    {
      typename image_traits<_KernelType>::intermediate_type sum;
      const int OffsetWidth = _SubWidth / 2;
      const int OffsetHeight = _SubHeight / 2;
      for (int y = 0, h = in.height(); y < h; ++y)
      {
        for (int x = 0, w = in.width(); x < w; ++x)
        {
          sum = 0;
          for (int yp = y - OffsetHeight, i = 0; yp <= y + OffsetHeight; ++yp)
          {
            for (int xp = x - OffsetWidth; xp <= x + OffsetWidth; ++xp, ++i)
            {
              if ((xp < w) && (xp >= 0) && (yp < h) && (yp >= 0))
              {
                sum += kernel_[i] * in.get(xp, yp);
              }
            }
          }
          out.set(x, y) = static_cast<_Tp>(sum);
        }
      }
    }
    _KernelType *kernel_;
  };

  // using subwindows
  template<size_t _SubWidth, size_t _SubHeight, typename _Tp, typename _KernelIterator>
    struct Convolution2
  {
    Convolution2(_KernelIterator kernel) : kernel_(kernel)
    {
    }
    template<typename SubWindowIterator> inline _Tp
      operator()(SubWindowIterator &in)
    {
      typename std::iterator_traits<_KernelIterator>::value_type tmp = 0;
      return static_cast<_Tp>(std::inner_product(in.begin(), in.end(), kernel_, tmp));
    }
    _KernelIterator kernel_;
  };

  // using proxy
  template<size_t _SubWidth, size_t _SubHeight, typename _Tp, typename _KernelIterator>
    struct Convolution3
  {
    Convolution3(_KernelIterator kernel) : kernel_(kernel)
    {
    }
    template<typename SubWindowIterator> inline _Tp
      operator()(SubWindowIterator &in)
    {
      return runHelper(in, std::iterator_traits<SubWindowIterator>::iterator_category());
    }

    // helper dispatch
    // for light (pointer-like) iterators
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
    runHelper(SubWindowIterator &in, std::forward_iterator_tag)
    {
      typename std::iterator_traits<_KernelIterator>::value_type tmp = 0;
      return static_cast<_Tp>(std::inner_product(in.begin(), in.end(), kernel_, tmp));
    }

    // helper dispatch
    // for heavyweight iterators 
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
    runHelper(SubWindowIterator &in, cuj::heavy_forward_iterator_tag)
    {
      // call std::copy on indirect iterators
      typename std::iterator_traits<_KernelIterator>::value_type tmp = 0;
      return static_cast<_Tp>(std::inner_product(&in.begin(), &in.end(), kernel_, tmp));
    }
    _KernelIterator kernel_;
  };


}

#endif
