
///////////////////////////////////////////////////////////////////////////////
//    Charles River Analytics Inc.
//    Copyright 2002-2003. All Rights Reserved.
//
//    ORIGINAL SYSTEM: The Vision Kit Library
//    SUBSYSTEM: Image
//    $Workfile: Cuj.cpp $
//    $Revision: 2 $
//    Author(s): V.Ablavsky, J.Pollak, M.Stevens
//
//    Description
//    ===========
//
// Currently demonstrates naive image processing operators
//
///////////////////////////////////////////////////////////////////////////////


#include "Image.h"
#include "ImageDefs.h"
#include "PortableGraymap.h"
#include "MedianFiltering.h"
#include "GammaCorrection.h"
#include "Window.h"
#include "Convolution.h"
#include <iostream>
#include <memory>
#include <ctime>
#include <algorithm>

inline double msecDeltaT(clock_t t1, clock_t t2)
{
  return 1e3*(static_cast<double>(t2-t1)/CLOCKS_PER_SEC);
}

// MedianFilteringTest
static inline void MedianFilterTest()
{
  clock_t tStart;
  clock_t tEnd;

#define LOAD_IMAGE
#ifdef LOAD_IMAGE 
  // load
  std::auto_ptr< cuj::Image<cuj::byte> > 
    in(cuj::PortableGraymap::read("Speckle.pgm"));
#else
  std::auto_ptr<cuj::Image<cuj::byte> > 
    in(new cuj::Image<cuj::byte> (1024, 1024));
#endif

    cuj::Image<cuj::byte> out(in->width(), in->height());

    // naive median filtering
    cuj::MedianFiltering1<3,3, cuj::byte> medianFiltering1;
    tStart = clock();
    medianFiltering1(*in, out);
    tEnd = clock();
    cuj::PortableGraymap::write("Speckle-1.pgm", out);
    std::cout << "\nMedianFiltering1(\"Speckle.pgm\")\n\tTime: " 
              << msecDeltaT(tStart, tEnd)
              << "ms\n\tSaving output to Speckle-1.pgm\n";

   // std::compliant approach
    cuj::MedianFiltering2<3,3, cuj::byte> medianFiltering2;
    tStart = clock();
    cuj::Window<3,3, cuj::byte> window(in->begin(), in->width(), in->height());
    std::transform(window.begin(),
                   window.end(), 
                   out.begin(),
                   medianFiltering2);
    tEnd = clock();
    cuj::PortableGraymap::write("Speckle-2.pgm", out);
    std::cout << "\nMedianFiltering2(\"Speckle.pgm\")\n\tTime: " 
              << msecDeltaT(tStart, tEnd)
              << "ms\n\tSaving output to Speckle-2.pgm\n";

    // another std::compliant approach
    cuj::MedianFiltering3<3,3, cuj::byte> medianFiltering3;
    tStart = clock();
    std::transform(window.begin(),
                   window.end(), 
                   out.begin(),
                   medianFiltering3);
    tEnd = clock();
    cuj::PortableGraymap::write("Speckle-3.pgm", out);
    std::cout << "\nMedianFiltering3(\"Speckle.pgm\")\n\tTime: " 
              << msecDeltaT(tStart, tEnd)
              << "ms\n\tSaving output to Speckle-3.pgm\n";


}

// Unsharp masking test
static inline void UnSharpMaskingTest()
{

  // load
#ifdef LOAD_IMAGE 
  std::auto_ptr< cuj::Image<cuj::byte> > 
    in(cuj::PortableGraymap::read("Blur.pgm"));
#else
  std::auto_ptr<cuj::Image<cuj::byte> > 
    in(new cuj::Image<cuj::byte> (1024, 1024));
#endif

  cuj::Image<cuj::byte> out1(in->width(), in->height());
  cuj::Image<cuj::byte> out2(in->width(), in->height());

  // sharpening kernel
  float kernel[9] = {-1.0f/9.0f, -1.0f/9.0f, -1.0f/9.0f,
                     -1.0f/9.0f,  8.0f/9.0f, -1.0f/9.0f,
                     -1.0f/9.0f, -1.0f/9.0f, -1.0f/9.0f};

  // naive unsharp masking
  cuj::Convolution1<3,3, cuj::byte, float> convolution1(&(kernel[0]));
  clock_t tStart = clock();
  convolution1(*in, out1);
  clock_t tEnd = clock();
  cuj::PortableGraymap::write("Blur-1.pgm", out1);
  std::cout << "\nUnSharpMasking1(\"Blur.pgm\")\n\tTime: " 
            << msecDeltaT(tStart, tEnd)
            << "ms\n\tSaving output to Blur-1.pgm\n";
  
  // std::compliant approach
  cuj::Convolution2<3,3, cuj::byte, float*> convolution2(&(kernel[0]));
  cuj::Window<3,3, cuj::byte> window(in->begin(), in->width(), in->height());
  tStart = clock();
  std::transform(window.begin(),
                 window.end(), 
                 out2.begin(), 
                 convolution2);
  tEnd = clock();
  cuj::PortableGraymap::write("Blur-2.pgm", out2);
  std::cout << "\nUnSharpMasking2(\"Blur.pgm\")\n\tTime: " 
            << msecDeltaT(tStart, tEnd)
            << "ms\n\tSaving output to Blur-2.pgm\n";

  // std::compliant approach using proxies
  cuj::Convolution3<3,3, cuj::byte, float*> convolution3(&(kernel[0]));
  tStart = clock();
  std::transform(window.begin(),
                 window.end(), 
                 out2.begin(), 
                 convolution3);
  tEnd = clock();
  cuj::PortableGraymap::write("Blur-3.pgm", out2);
  std::cout << "\nUnSharpMasking3(\"Blur.pgm\")\n\tTime: " 
            << msecDeltaT(tStart, tEnd)
            << "ms\n\tSaving output to Blur-3.pgm\n";

}

static inline void GammaCorrectionTest()
{

  // the gamma value
  float gamma = 1.7f;

#ifdef LOAD_IMAGE 
  std::auto_ptr< cuj::Image<cuj::byte> > 
    in(cuj::PortableGraymap::read("Dark.pgm"));
#else
  std::auto_ptr<cuj::Image<cuj::byte> > 
    in(new cuj::Image<cuj::byte> (1024, 1024));
#endif
  cuj::Image<cuj::byte> out1(in->width(), in->height());
  cuj::Image<cuj::byte> out2(in->width(), in->height());
  
  // naive gamma correction
  cuj::GammaCorrection1<cuj::byte> gammaCorrection1(gamma);
  clock_t tStart = clock();
  gammaCorrection1(*in, out1);
  clock_t tEnd = clock();
  cuj::PortableGraymap::write("Dark-1.pgm", out1);
  std::cout << "\nGammaCorrection1(\"Dark.pgm\")\n\tTime: " 
            << msecDeltaT(tStart, tEnd)
            << "ms\n\tSaving output to Dark-1.pgm\n";
  
  // std::compliant gamma correction
  cuj::GammaCorrection2<cuj::byte> gammaCorrection2(gamma);
  tStart = clock();
  std::transform(in->begin(), 
                 in->end(),
                 out2.begin(),
                 gammaCorrection2);
  tEnd = clock();
  cuj::PortableGraymap::write("Dark-2.pgm", out1);
  std::cout << "\nGammaCorrection2(\"Dark.pgm\")\n\tTime: " 
            << msecDeltaT(tStart, tEnd)
            << "ms\n\tSaving output to Dark-2.pgm\n";
  
}

int main(int argc, char* argv[])
{

  // check arguments and load image
  if (argc != 1)
  {
    std::cerr << "Usage: " << argv[0] << "\n";
    return 1;
  }

  try
  {

    MedianFilterTest();
    UnSharpMaskingTest();
    GammaCorrectionTest();

  
    // success
    return 0;

  }
  catch (std::exception &e)
  {
    std::cerr << "Caught exception: " << e.what() << "\n";
  }

  // default is failed
  return 1;
}
