///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    DynSystemModule.cpp
// Package:     SDESolver
// Purpose:     Load a shared object module and get pointer to the CDynSystem instance
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <dlfcn.h>
#include "DynSystemModule.h"
#include "Util.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CDynSystemModule
// Class:       CDynSystemModule
// Purpose:     Default constructor
// Inputs:      strDynSystem        name of dynamical system e.g. EllipseDynSystem
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CDynSystemModule::CDynSystemModule(std::string strDynSystem)
: m_strDynSystem(strDynSystem), m_pDynSystem(NULL), m_pModule(NULL)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CDynSystemModule
// Class:       CDynSystemModule
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CDynSystemModule::~CDynSystemModule(void)
{
    Unload();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDynSystem
// Class:       CDynSystemModule
// Purpose:     Get pointer to dynamical system. Lazy loads.
// Inputs:      None
// Outputs:     returns         pointer to dynamical system or NULL if an error occurred
///////////////////////////////////////////////////////////////////////////////////////////////
CDynSystem* CDynSystemModule::GetDynSystem(void)
{
    if (NULL == m_pDynSystem) Load();
    return m_pDynSystem;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Load
// Class:       CDynSystemModule
// Purpose:     Load shared library
// Inputs:      None
// Outputs:     returns         true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CDynSystemModule::Load(void)
{
    // Prepare name of shared library
    char* strModule = (char*)xmalloc(m_strDynSystem.length() + 6);
    sprintf(strModule, "./%s.so", m_strDynSystem.c_str());

    // Open shared library
    m_pModule = dlopen(strModule, RTLD_LAZY);
    free(strModule);
    if (NULL == m_pModule)
    {
        fprintf(stderr, "%s\n", dlerror());
        return false;
    }

    // Get pointer to "create" function
    Create_t create = (Create_t)dlsym(m_pModule, "Create");
    if (NULL == create)
    {
        fprintf(stderr, "%s\n", dlerror());
        return false;
    }

    // Create object
    m_pDynSystem = create();
    if (NULL == m_pDynSystem) return false;

    // Success
    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Unload
// Class:       CDynSystemModule
// Purpose:     Unload shared library
// Inputs:      None
// Outputs:     returns         true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CDynSystemModule::Unload(void)
{
    // Sanity check
    if ((NULL == m_pModule) || (NULL == m_pDynSystem)) return false;

    // Get pointer to "destroy" function
    Destroy_t destroy = (Destroy_t)dlsym(m_pModule, "Destroy");
    if (NULL == destroy)
    {
        fprintf(stderr, "%s\n", dlerror());
        return false;
    }

    // Destroy object
    destroy(m_pDynSystem);
}
