///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    EMSolver.cpp
// Package:     SDESolver
// Class:       CEMSolver
// Purpose:     Euler-Maruyama SDE solver. Strong order 0.5
//              Ref: Kloeden, P. E. et al (1999): "Numerical Solution of Stochastic
//              Differential Equations", Springer-Verlag.
// Written:     Dave Burke      03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include "EMSolver.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CEMSolver
// Class:       CEMSolver
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CEMSolver::CEMSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CEMSolver
// Class:       CEMSolver
// Purpose:     Parameterised constructor
// Inputs:      rSettings       settings for the solver
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CEMSolver::CEMSolver(const CSolverSettings& rSettings)
: CSolver(rSettings)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CEMSolver
// Class:       CEMSolver
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CEMSolver::~CEMSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Advance
// Class:       CEMSolver
// Purpose:     Advance the solution
// Inputs:      nTime               the current time
//              ndt                 time step
//              pdW                 dW
//              pdZ                 dZ
//              pXin                X at current time
// Outputs:     pXout               X advanced by ndt
//              returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CEMSolver::Advance(float nTime, float ndt, float* pdW,
                        float* pdZ, float* pXin, float* pXout)
{
    float nB;

    // For each dimension, iterate the solution by ndt
    for (unsigned int i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        // Calculate b by matrix multiplication
        nB = 0;
        for (unsigned int k = 0; k < m_pDynSystem->GetDimension(); ++k)
        {
            nB += m_ppB[i][k]*pdW[k];
        }

        pXout[i] = pXin[i] + m_pA[i]*ndt + nB;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDescription
// Class:       CEMSolver
// Purpose:     Get a description of the solver. Subclasses should override this method.
// Inputs:      None
// Outputs:     returns                a short description of the solver
///////////////////////////////////////////////////////////////////////////////////////////////
std::string CEMSolver::GetDescription(void)
{
    return "Euler-Maruyama solver (explicit). Strong order 0.5.";
}

