///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    EllipseDynSystem.cpp
// Package:     SDESolver
// Class:       CDynSystem
// Purpose:     SDE system in Ito form:
//                  dX1 = [dH/dX2 - l*(H)*dH/dX1]dt + [-l*sigma*dH/dX1]dW
//                  dX2 = [-dH/dX1 - l*(H)*dH/dX2]dt + [-l*sigma*dH/dX2]dW              ...1
//
//              where W is Wiener process and formally dW/dt = n(t) white noise, and
//              H is the equation for the Ellipse:
//                  X1^2/a^2 + X2^2/b^2 - 1                                             ...2
//              and l > 0, sigma is noise intensity
//
// Written:     Dave Burke      03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "EllipseDynSystem.h"

#define DIM 2	// dimension 2

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CEllipseDynSystem
// Class:       CEllipseDynSystem
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CEllipseDynSystem::CEllipseDynSystem(void)
{
    // Register our dimension
    m_nDimension = DIM;

    // System parameters
    m_nA = 2;
    m_nB = 1;
    m_nL = 0.08;
    m_nS = 0.5;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDriftAndDiffusion
// Class:       CEllipseDynSystem
// Purpose:     Get the drift vector (a) and diffusion matrix (b) for the system
// Inputs:      nTime           time
//              pXin            X vector at this time
// Outputs:     pAout           a vector
//              pBout           b matrix
///////////////////////////////////////////////////////////////////////////////////////////////
void CEllipseDynSystem::GetDriftAndDiffusion(float nTime, float* pXin,
                                             float* pAout, float** ppBout)
{
    float X1 = pXin[0];
    float X2 = pXin[1];
    float nH, nH1, nH2;

    nH = pow(X1, 2)/pow(m_nA, 2) + pow(X2, 2)/pow(m_nB, 2) - 1;
    nH1 = 2*X1/pow(m_nA,2);
    nH2 = 2*X2/pow(m_nB,2);

    // SDE a vector
    pAout[0] = nH2 - m_nL*nH*nH1;
    pAout[1] = -nH1 - m_nL*nH*nH2;

    // SDE b matrix
    ppBout[0][0] = m_nS*m_nL*nH1;
    ppBout[0][1] = 0;
    ppBout[1][0] = 0;
    ppBout[1][1] = m_nS*m_nL*nH2;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Create
// Purpose:     Create an instance of object. Used in shared libraries
// Inputs:      None
// Outputs:     return          pointer to new instance
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" CDynSystem* Create(void)
{
    return new CEllipseDynSystem();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Destroy
// Purpose:     Destroy an instance of object. Used in shared libraries
// Inputs:      pDynSystem      object to destroy
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" void Destroy(CDynSystem* pDynSystem)
{
    delete pDynSystem;
}


