///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    GBMDynSystem.cpp
// Package:     SDESolver
// Class:       CDynSystem
// Purpose:     SDE system in Ito form:
//                  dX = mu*X*dt + sigma*X*dW                                           ...1
//              where W is Wiener process and formally dW/dt = n(t) is white noise.
//              This is Geometric Brownian Motion and has explicit solution of
//              X = X(0)exp((mu-0.5*sigma^2)*(t+sigma*W))
//
// Written:        Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "GBMDynSystem.h"

#define DIM 1    // dimension 1

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CGBMDynSystem
// Class:       CGBMDynSystem
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CGBMDynSystem::CGBMDynSystem(void)
{
    // Register our dimension
    m_nDimension = DIM;

    // System parameters
    m_nMu = 1;
    m_nSigma = 0.6;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDriftAndDiffusion
// Class:       CGBMDynSystem
// Purpose:     Get the drift vector (a) and diffusion matrix (b) for the system
// Inputs:      nTime               time
//              pXin                X vector at this time
// Outputs:     pAout               a vector
//              pBout               b matrix
///////////////////////////////////////////////////////////////////////////////////////////////
void CGBMDynSystem::GetDriftAndDiffusion(float nTime, float* pXin,
                                         float* pAout, float** ppBout)
{
    float X = pXin[0];

    // SDE a vector
    pAout[0] = m_nMu*X;
    ppBout[0][0] = m_nSigma*X;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Create
// Purpose:     Create an instance of object. Used in shared libraries
// Inputs:      None
// Outputs:     return            pointer to new instance
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" CDynSystem* Create(void)
{
    return new CGBMDynSystem();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Destroy
// Purpose:     Destroy an instance of object. Used in shared libraries
// Inputs:      pDynSystem        object to destroy
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" void Destroy(CDynSystem* pDynSystem)
{
    delete pDynSystem;
}

