///////////////////////////////////////////////////////////////////////////////
//    Charles River Analytics Inc.
//    Copyright 2002-2003. All Rights Reserved.
//
//    ORIGINAL SYSTEM: The Vision Kit Library
//    SUBSYSTEM: Image
//    $Workfile: MedianFiltering.h $
//    $Revision: 2 $
//    Author(s): V.Ablavsky, J.Pollak, M.Stevens
//
//    Description
//    ===========
//
// Implements the naive, 4-for median filtering algorithm
//
///////////////////////////////////////////////////////////////////////////////

#ifndef MedianFiltering_h
#define MedianFiltering_h

#include "Image.h"
#include "SubWindow.h"
#include <vector>

namespace cuj
{

  // Four nested loops
  template<size_t _SubWidth, size_t _SubHeight, typename _Tp>  struct MedianFiltering1
  {
    MedianFiltering1() : data(_SubWidth * _SubHeight)
    {
    }
    inline void operator()(const cuj::Image<_Tp> &in, cuj::Image<_Tp> &out)
    {
      const int OffsetWidth = _SubWidth / 2;
      const int OffsetHeight = _SubHeight / 2;
      for (int y = 0, h = in.height(); y < h; ++y)
      {
        for (int x = 0, w = in.width(); x < w; ++x)
        {
          for (int yp = y - OffsetHeight, i = 0; yp <= y + OffsetHeight; ++yp)
          {
            for (int xp = x - OffsetWidth; xp <= x + OffsetWidth; ++xp, ++i)
            {
              if ((xp < w) && (xp >= 0) && (yp < h) && (yp >= 0))
              {
                data[i] = in.get(xp, yp);
              }
              else
              {
                data[i] = 0;
              }
            }
          }
          std::nth_element(data.begin(), data.begin() + (_SubWidth * _SubHeight) / 2, data.end());
          out.set(x, y) = data[(_SubWidth * _SubHeight) / 2];
        }
      }
    }    
    std::vector<_Tp> data;

  };


  template<size_t _SubWidth, size_t _SubHeight, typename _Tp> struct MedianFiltering2
  {

    MedianFiltering2() : data(_SubWidth * _SubHeight)
    {
    }
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
      operator()(SubWindowIterator &in)
    {
      std::copy(in.begin(), in.end(), data.begin());
      std::nth_element(data.begin(), data.begin() + (_SubWidth * _SubHeight) / 2, data.end());
      return data[(_SubWidth * _SubHeight) / 2];
    }
    std::vector<_Tp> data;
  };


  template<size_t _SubWidth, size_t _SubHeight, typename _Tp> struct MedianFiltering3
  {

    // construct
    MedianFiltering3() : data(_SubWidth * _SubHeight)
    {
    }

    // process
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
      operator()(SubWindowIterator &in)
    {
      return runHelper(in, std::iterator_traits<SubWindowIterator>::iterator_category());
    }

    // helper dispatch
    // for light (pointer-like) iterators
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
    runHelper(SubWindowIterator &in, std::forward_iterator_tag)
    {
      std::copy(in.begin(), in.end(), data.begin());
      std::nth_element(data.begin(), data.begin() + (_SubWidth * _SubHeight) / 2, data.end());
      return data[(_SubWidth * _SubHeight) / 2];
    }

    // helper dispatch
    // for heavyweight iterators 
    template<typename SubWindowIterator> inline 
      typename std::iterator_traits<SubWindowIterator>::value_type
    runHelper(SubWindowIterator &in, cuj::heavy_forward_iterator_tag)
    {
      // call std::copy on indirect iterators
      std::copy(&in.begin(), &in.end(), data.begin());
      std::nth_element(data.begin(), data.begin() + (_SubWidth * _SubHeight) / 2, data.end());
      return data[(_SubWidth * _SubHeight) / 2];
    }
    std::vector<_Tp> data;
  };


}

#endif
