///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    MilsteinSolver.cpp
// Package:     SDESolver
// Class:       CMilsteinSolver
// Purpose:     Milstein SDE solver. For simplicity we assume diagonal noise i.e.
//              bjk==0 for j!=k. Strong order 1.0. Avoids use of derivatives (i.e. Runge-Kutta
//              style approach).
//              Ref: Kloeden, P. E. et al (1999): "Numerical Solution of Stochastic
//              Differential Equations", Springer-Verlag.
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "MilsteinSolver.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CMilsteinSolver
// Class:       CMilsteinSolver
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CMilsteinSolver::CMilsteinSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CMilsteinSolver
// Class:       CMilsteinSolver
// Purpose:     Parameterised constructor
// Inputs:      rSettings           settings for the solver
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CMilsteinSolver::CMilsteinSolver(const CSolverSettings& rSettings)
: CSolver(rSettings)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CMilsteinSolver
// Class:       CMilsteinSolver
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CMilsteinSolver::~CMilsteinSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Initialise
// Class:       CMilsteinSolver
// Purpose:     Create memory for supporting vectors
// Inputs:      None
// Outputs:     returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CMilsteinSolver::Initialise(void)
{
    // Call base class initialisation
    if (!CSolver::Initialise()) return false;

    m_pSupportX = MakeVector(m_pDynSystem->GetDimension());
    m_pSupportA = MakeVector(m_pDynSystem->GetDimension());
    m_ppSupportB = MakeMatrix(m_pDynSystem->GetDimension(), m_pDynSystem->GetDimension());

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Advance
// Class:       CMilsteinSolver
// Purpose:     Advance the solution
// Inputs:      nTime               the current time
//              ndt                 time step
//              pdW                 dW
//              pdZ                 dZ
//              pXin                X at current time
// Outputs:     pXout               X advanced by ndt
//              returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CMilsteinSolver::Advance(float nTime, float ndt, float* pdW,
                              float* pdZ, float* pXin, float* pXout)
{
    unsigned int i, j;

    // Check we have diagonal noise
    if (!CheckDiagonalNoise())
    {
        m_strLastError = "Only diagonal noise (bjk == 0, j!=k) allowed";
        return false;
    }

    // For each dimension, iterate the solution by ndt
    for (i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        // Vector supporting value
        for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
        {
            m_pSupportX[j] = pXin[j] + m_pA[j]*ndt + m_ppB[j][i]*sqrt(ndt);
        }

        m_pDynSystem->GetDriftAndDiffusion(nTime, m_pSupportX, m_pSupportA, m_ppSupportB);

        pXout[i] = pXin[i] + m_pA[i]*ndt + m_ppB[i][i]*pdW[i] + 1.0/(2.0*sqrt(ndt))
                 *(m_ppSupportB[i][i] - m_ppB[i][i])*(pow(pdW[i], 2.0) - ndt);
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Finish
// Class:       CMilsteinSolver
// Purpose:     Write data (solution) to file. This method may be overriden to provide
//              solver-specific cleanup.
// Inputs:      None
// Outputs:     returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CMilsteinSolver::Finish(void)
{
    free(m_pSupportX);
    free(m_pSupportA);

    for (unsigned int i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        free(m_ppSupportB[i]);
    }
    free(m_ppSupportB);

    // Call base class finish
    return CSolver::Finish();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CheckDiagonalNoise
// Class:       CMilsteinSolver
// Purpose:     Check for diagonal noise
// Inputs:      None
// Outputs:     returns                 true if system has diagonal noise
///////////////////////////////////////////////////////////////////////////////////////////////
bool CMilsteinSolver::CheckDiagonalNoise(void)
{
    unsigned int i, j;

    for (i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
        {
            if ((i != j) && (m_ppB[i][j] != 0))
            {
                return false;
            }
        }
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDescription
// Class:       CMilsteinSolver
// Purpose:     Get a description of the solver. Subclasses should override this method.
// Inputs:      None
// Outputs:     returns                 a short description of the solver
///////////////////////////////////////////////////////////////////////////////////////////////
std::string CMilsteinSolver::GetDescription(void)
{
    return "Milstein solver (explicit). Strong order 1.0. Assumes diagonal noise i.e. bk==0 for j!=k. " \
           "Avoids use of derivatives (i.e. Runge-Kutta style approach)";
}


