///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    RungeKuttaSolver.cpp
// Package:     SDESolver
// Class:       CRungeKuttaSolver
// Purpose:     Runge-Kutta SDE solver. For simplicity we assume diagonal noise i.e.
//              bjk==0 for j!=k. Strong order 1.5. For autonomous systems.
//              Ref: Kloeden, P. E. et al (1999): "Numerical Solution of Stochastic
//              Differential Equations", Springer-Verlag.
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "RungeKuttaSolver.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CRungeKuttaSolver
// Class:       CRungeKuttaSolver
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CRungeKuttaSolver::CRungeKuttaSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CRungeKuttaSolver
// Class:       CRungeKuttaSolver
// Purpose:     Parameterised constructor
// Inputs:      rSettings           settings for the solver
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CRungeKuttaSolver::CRungeKuttaSolver(const CSolverSettings& rSettings)
: CSolver(rSettings)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CRungeKuttaSolver
// Class:       CRungeKuttaSolver
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CRungeKuttaSolver::~CRungeKuttaSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Initialise
// Class:       CRungeKuttaSolver
// Purpose:     Create memory for supporting vectors
// Inputs:      None
// Outputs:     returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CRungeKuttaSolver::Initialise(void)
{
    // Call base class initialisation
    if (!CSolver::Initialise()) return false;

    unsigned int d = m_pDynSystem->GetDimension();

    m_pS1 = (SUPPORT*)malloc(sizeof(SUPPORT) * d);
    m_pS2 = (SUPPORT*)malloc(sizeof(SUPPORT) * d);
    m_pS3 = (SUPPORT*)malloc(sizeof(SUPPORT) * d);
    m_pS4 = (SUPPORT*)malloc(sizeof(SUPPORT) * d);

    for (unsigned int i = 0; i < d; ++i)
    {
        m_pS1[i].m_pX = MakeVector(d);
        m_pS1[i].m_pA = MakeVector(d);
        m_pS1[i].m_ppB = MakeMatrix(d, d);

        m_pS2[i].m_pX = MakeVector(d);
        m_pS2[i].m_pA = MakeVector(d);
        m_pS2[i].m_ppB = MakeMatrix(d, d);

        m_pS3[i].m_pX = MakeVector(d);
        m_pS3[i].m_pA = MakeVector(d);
        m_pS3[i].m_ppB = MakeMatrix(d, d);

        m_pS4[i].m_pX = MakeVector(d);
        m_pS4[i].m_pA = MakeVector(d);
        m_pS4[i].m_ppB = MakeMatrix(d, d);
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Advance
// Class:       CRungeKuttaSolver
// Purpose:     Advance the solution
// Inputs:      nTime               the current time
//              ndt                 time step
//              pdW                 dW
//              pdZ                 dZ
//              pXin                X at current time
// Outputs:     pXout               X advanced by ndt
//              returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CRungeKuttaSolver::Advance(float nTime, float ndt, float* pdW,
                                float* pdZ, float* pXin, float* pXout)
{
    unsigned int i, j;
    float nTemp(0);
    unsigned d = m_pDynSystem->GetDimension();
    float m = (float)d;

    // Check we have diagonal noise
    if (!CheckDiagonalNoise())
    {
        m_strLastError = "Only diagonal noise (bjk == 0, j!=k) allowed";
        return false;
    }

    // Create supporting vectors
    for (i = 0; i < d; ++i)
    {
        for (j = 0; j < d; ++j)
        {
            nTemp = pXin[j] + 1/m*m_pA[j]*ndt;
            m_pS1[i].m_pX[j] = nTemp + m_ppB[j][i]*sqrt(ndt);
            m_pS2[i].m_pX[j] = nTemp - m_ppB[j][i]*sqrt(ndt);
        }

        // nTime irrelevant (autonomous) here
        m_pDynSystem->GetDriftAndDiffusion(nTime,
                                           m_pS1[i].m_pX,
                                           m_pS1[i].m_pA,
                                           m_pS1[i].m_ppB);
        m_pDynSystem->GetDriftAndDiffusion(nTime,
                                           m_pS2[i].m_pX,
                                           m_pS2[i].m_pA,
                                           m_pS2[i].m_ppB);
        for (j = 0; j < d; ++j)
        {
            m_pS3[i].m_pX[j] = m_pS1[i].m_pX[j] + m_pS1[i].m_ppB[j][i]*sqrt(ndt);
            m_pS4[i].m_pX[j] = m_pS1[i].m_pX[j] - m_pS1[i].m_ppB[j][i]*sqrt(ndt);
        }

        // nTime irrelevant (autonomous) here
        m_pDynSystem->GetDriftAndDiffusion(nTime,
                                           m_pS3[i].m_pX,
                                           m_pS3[i].m_pA,
                                           m_pS3[i].m_ppB);
        m_pDynSystem->GetDriftAndDiffusion(nTime,
                                           m_pS4[i].m_pX,
                                           m_pS4[i].m_pA,
                                           m_pS4[i].m_ppB);
    }


    // For each dimension, iterate the solution by ndt.
    // For clarity we'll do this line for line as per Kloeden et al p380
    for (i = 0; i < d; ++i)
    {
        // 1
        pXout[i] = pXin[i] + m_ppB[i][i]*pdW[i];

        // 2
        nTemp = 0;
        for (j = 0; j < d; ++j)
        {
            nTemp += (m_pS1[j].m_pA[i] - m_pS2[j].m_pA[i])*pdZ[i];
        }
        nTemp *= 1.0/(2.0*sqrt(ndt));
        pXout[i] += nTemp;

        // 3
        nTemp = 0;
        for (j = 0; j < d; ++j)
        {
            nTemp += m_pS1[j].m_pA[i] - 2.0*(m-2.0)/m*m_pA[i] + m_pS2[j].m_pA[i];
        }
        nTemp *= (1.0/4.0)*ndt;
        pXout[i] += nTemp;

        // 4
        pXout[i] += 1.0/(4.0*sqrt(ndt)) * (m_pS1[i].m_ppB[i][i] - m_pS2[i].m_ppB[i][i])*(pow(pdW[i], 2.0) - ndt);

        // 5
        pXout[i] += 1.0/(2.0*ndt) * (m_pS1[i].m_ppB[i][i] - 2.0*m_ppB[i][i] + m_pS2[i].m_ppB[i][i])*(pdW[i]*ndt - pdZ[i]);

        // 6
        pXout[i] += 1.0/(4.0*ndt) * (m_pS3[i].m_ppB[i][i] - m_pS4[i].m_ppB[i][i] - m_pS1[i].m_ppB[i][i] + m_pS2[i].m_ppB[i][i])
                  * (1.0/3.0*pow(pdW[i], 2.0) - ndt)*pdW[i];
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Finish
// Class:       CRungeKuttaSolver
// Purpose:     Write data (solution) to file. This method may be overriden to provide
//              solver-specific cleanup.
// Inputs:      None
// Outputs:     returns                true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CRungeKuttaSolver::Finish(void)
{
    for (unsigned int i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        free(m_pS1[i].m_pX);
        free(m_pS1[i].m_pA);
        free(m_pS2[i].m_pX);
        free(m_pS2[i].m_pA);
        free(m_pS3[i].m_pX);
        free(m_pS3[i].m_pA);
        free(m_pS4[i].m_pX);
        free(m_pS4[i].m_pA);

        for (unsigned int j = 0; j < m_pDynSystem->GetDimension(); ++j)
        {
            free(m_pS1[i].m_ppB[j]);
            free(m_pS2[i].m_ppB[j]);
            free(m_pS3[i].m_ppB[j]);
            free(m_pS4[i].m_ppB[j]);
        }
        free(m_pS1[i].m_ppB);
        free(m_pS2[i].m_ppB);
        free(m_pS3[i].m_ppB);
        free(m_pS4[i].m_ppB);
    }
    free(m_pS1);
    free(m_pS2);
    free(m_pS3);
    free(m_pS4);

    // Call base class finish
    return CSolver::Finish();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CheckDiagonalNoise
// Class:       CRungeKuttaSolver
// Purpose:     Check for diagonal noise
// Inputs:      None
// Outputs:     returns                true if system has diagonal noise
///////////////////////////////////////////////////////////////////////////////////////////////
bool CRungeKuttaSolver::CheckDiagonalNoise(void)
{
    unsigned int i, j;

    for (i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
        {
            if ((i != j) && (m_ppB[i][j] != 0))
            {
                return false;
            }
        }
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDescription
// Class:       CRungeKuttaSolver
// Purpose:     Get a description of the solver. Subclasses should override this method.
// Inputs:      None
// Outputs:     returns                a short description of the solver
///////////////////////////////////////////////////////////////////////////////////////////////
std::string CRungeKuttaSolver::GetDescription(void)
{
    return "Runge-Kutta solver (explicit). Strong order 1.5. Assumes diagonal noise i.e. " \
           "bjk==0 for j!=k. For autonomous systems";
}



