///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    SDESolver.cpp
// Package:     SDESolver
// Purpose:     Toolbox for numerical solution of Stochastic Differential Equations (SDE).
//              Extensible framework allows for additional solvers to be included in the
//              toolbox. SDE systems are specified as shared object modules.
//              Current supported solvers:
//                  - Euler-Maruyama 0.5
//                  - Milstein 1.0
//                  - Runge-Kutta 1.5
// Written:        Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Util.h"
#include "SolverSettings.h"
#include "SolverFactory.h"
#include "Solver.h"
#include "DynSystem.h"
#include "DynSystemModule.h"

/* The name of this program. */
const char* g_strProgramName;

/* Prints usage information for this program to STREAM (typically
   stdout or stderr), and exit the program with EXIT_CODE.  Does not
   return.  */
void PrintUsage (FILE* stream, int exit_code)
{
    fprintf (stream, "Numerical solver for SDEs.\nDave Burke 2002\nUsage: %s options \n", g_strProgramName);
    fprintf (stream, "  -h  --help          Display this usage information.\n"
                     "  -s  --solver        Solver to use: EM, MILSTEIN, or RUNGE_KUTTA\n"
                     "  -d  --dynsystem     Dynamical system name\n"
                     "  -n  --stepsize      Stepsize for Wiener increments\n"
                     "  -m  --multiplier    Solver step of multipler*stepsize\n"
                     "  -t  --time          Solve from 0 to time\n"
                     "  -i  --ics           Comma delimited initial conditions\n"
                     "  -o  --output        Output file for solution\n"
                     "  -r  --randomseed    Optional integer for seeding random number generator\n"
                     "  -w  --wieneroutput  Optional filename to write Wiener realisation\n");
      exit (exit_code);
}

int main (int argc, char* argv[])
{
    // Program name
    g_strProgramName = argv[0];

    // A string listing valid short options letters.
     const char* const short_options = "hs:d:n:m:t:i:o:r:w:";

    // An array describing valid long options.
    const struct option long_options[] =
    {
        { "help",           0, NULL, 'h' },
        { "solver",         0, NULL, 's' },
        { "dynsystem",      0, NULL, 'd' },
        { "stepsize",       0, NULL, 'n' },
        { "multiplier",     0, NULL, 'm' },
        { "time",           0, NULL, 't' },
        { "ics",            0, NULL, 'i' },
        { "output",         0, NULL, 'o' },
        { "randomseed",     0, NULL, 'r' },
        { "writewiener",    0, NULL, 'w' },
        { NULL,             0, NULL, 0   }   /* Required at end of array.  */
    };

    // Get CLAs
    CSolverSettings settings;
    char* strSolverName = NULL;
    char* strDynSystem = NULL;
    char* strICs = NULL;

    int nNextOption;
    do
    {
        nNextOption = getopt_long (argc, argv, short_options,
                                   long_options, NULL);
        switch (nNextOption)
        {
            case 'h':   /* -h or --help */
                PrintUsage(stdout, 0);
                break;

            case 's':   /* -s or --solver */
                 strSolverName = optarg;
                break;

            case 'd':   /* -s or --solver */
                strDynSystem = optarg;
                break;

            case 'n':   /* -n or --stepsize */
                settings.m_nStepSize = atof(optarg);
                break;

            case 'm':   /* -m or --multiplier */
                settings.m_nMultiplier = atoi(optarg);
                break;

            case 't':   /* -t or --time */
                settings.m_nT = atof(optarg);
                break;

            case 'i':   /* -i or --ics */
                strICs = optarg;
                break;

            case 'o':   /* -o or --output */
                settings.m_strOutputFileName = optarg;
                break;

            case 'r':   /* -r or --randomseed */
                settings.m_nSeed = atol(optarg);
                break;

            case 'w':   /* -w or --writewiener */
                settings.m_strWienerOutputFileName = optarg;
                break;

            case '?':   /* The user specified an invalid option.  */
                PrintUsage(stderr, 1);

            case -1:    /* Done with options.  */
                break;

            default:    /* Something else: unexpected.  */
                abort();
        }
    }
    while (nNextOption != -1);

    // Sanity check
    if ((NULL == strSolverName) ||
        (NULL == strDynSystem) ||
        (NULL == strICs))
    {
        PrintUsage(stderr, 1);
        return 0;
    }

    // Load the dynamical system shared object
    CDynSystem* pDynSystem = NULL;
    CDynSystemModule module(strDynSystem);
    pDynSystem = module.GetDynSystem();
    if (NULL == pDynSystem)
    {
        fprintf(stderr, "Unable to load shared library module ./%s.so\n", strDynSystem);
        exit(1);
    }

    // Prepare initial conditions
    settings.m_ICs.m_nDim = pDynSystem->GetDimension();
    settings.m_ICs.m_pData = (float*)xcalloc(settings.m_ICs.m_nDim, sizeof(float));
    char strSeps[] = ",";
    char* strTemp = strdup(strICs);
    char* strToken = strtok(strTemp, strSeps);

    int n(0);
    while (strToken != NULL)
    {
        // Make sure we don't get too many!
        if (n >= settings.m_ICs.m_nDim) break;
        settings.m_ICs.m_pData[n] = atof(strToken);

        strToken = strtok(NULL, strSeps);
        n++;
    }
    free (strTemp);

    // Create solver
    CSolver* pSolver = CSolverFactory::CreateSolver(strSolverName, settings);
    if (NULL == pSolver)
    {
        fprintf(stderr, "Unable to create solver with name %s\n", strSolverName);
        exit(1);
    }

    // Run
    pSolver->RegisterDynSystem(pDynSystem);
    if (!pSolver->Run())
    {
        fprintf(stderr, "%s\n", pSolver->GetLastError().c_str());
        delete pSolver;
        exit(1);
    }

    // Cleanup
    delete pSolver;

    // Success
    return 0;
}
