///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    Solver.cpp
// Package:     SDESolver
// Class:       CSolver
// Purpose:     Base class for Stochastic Differential Equations (SDE) solvers. Uses the
//              Template pattern for defining the solution procedure
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <time.h>
#include <math.h>
#include "Solver.h"
#include "Util.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CSolver
// Class:       CSolver
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSolver::CSolver(void)
: m_ppdW(NULL), m_ppdZ(NULL), m_pDynSystem(NULL), m_ppSolution(NULL), m_nSolutionSize(0),
  m_nWienerSize(0), m_pA(NULL), m_ppB(NULL)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CSolver
// Class:       CSolver
// Purpose:     Parameterised constructor
// Inputs:      rSettings        settings for the solver
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSolver::CSolver(const CSolverSettings& rSettings)
: m_settings(rSettings), m_ppdW(NULL), m_ppdZ(NULL), m_pDynSystem(NULL), m_ppSolution(NULL),
  m_nSolutionSize(0), m_nWienerSize(0), m_pA(NULL), m_ppB(NULL)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CSolver
// Class:       CSolver
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSolver::~CSolver(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Cleanup
// Class:       CSolver
// Purpose:     Cleanup allocated memory
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
void CSolver::Cleanup(void)
{
    unsigned int i;

    if (NULL != m_ppSolution)
    {
        for (i = 0; i < m_nSolutionSize; ++i)
        {
            free(m_ppSolution[i]);
            m_ppSolution[i] = NULL;
        }
        free(m_ppSolution);
        m_ppSolution = NULL;
    }

    if (NULL != m_ppdW)
    {
        for (i = 0; i < m_nWienerSize; ++i)
        {
            free(m_ppdW[i]);
            m_ppdW[i] = NULL;
        }
        free(m_ppdW);
        m_ppdW = NULL;
    }

    if (NULL != m_ppdZ)
    {
        for (i = 0; i < m_nWienerSize; ++i)
        {
            free(m_ppdZ[i]);
            m_ppdZ[i] = NULL;
        }
        free(m_ppdZ);
        m_ppdZ = NULL;
    }

    if (NULL != m_pA)
    {
        free(m_pA);
        m_pA = NULL;
    }

    if (NULL != m_ppB)
    {
        for (i = 0; i < m_pDynSystem->GetDimension(); ++i)
        {
            free(m_ppB[i]);
            m_ppB[i] = NULL;
        }
        m_ppB = NULL;
    }

    // Don't delete m_pDynSystem since it might
    // have been created in shared object
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Initialise
// Class:       CSolver
// Purpose:     Create memory and discretised Brownian path. This method may be overriden to
//              provide solver-specific initialisation
// Inputs:      None
// Outputs:     returns                true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CSolver::Initialise(void)
{
    unsigned int i, j;

    printf("Initialising...\n");

    // Sanity check
    if (NULL == m_pDynSystem)
    {
        m_strLastError = "Dynamical system not registered to solver";
        return false;
    }

    // Sanity check
    if ((0 == m_settings.m_nT) ||
        (0 == m_settings.m_nStepSize) ||
        (0 == m_settings.m_nMultiplier))
    {
        m_strLastError = "Nothing to do";
        return false;
    }

    // Create memory to hold solution and Brownian path - round to nearest integer
    // (don't need Wiener increment at endpoint)
    m_nSolutionSize = (unsigned int)(m_settings.m_nT
                    / (m_settings.m_nStepSize*(float)m_settings.m_nMultiplier) + 0.5);

    m_nWienerSize = (unsigned int) (m_nSolutionSize * m_settings.m_nMultiplier -  m_settings.m_nMultiplier);

    // Create solution matrix
    m_ppSolution = MakeMatrix(m_nSolutionSize, m_pDynSystem->GetDimension()+1);

    // Set initial conditions - note we reserve the first column of m_ppSolution for time
    m_ppSolution[0][0] = 0;    // t = 0
    float* pICs = m_settings.m_ICs.m_pData;
    for (i = 0; i < m_pDynSystem->GetDimension(); ++i)
    {
        m_ppSolution[0][i+1] = (NULL != pICs) ? pICs[i] : 0;
    }

    // Create Wiener increment dW. Also create dZ, which is an approximation of the double
    // integral dWdt - see Kloeden et al. Note E[dZ]=0, E[dZ^2]=1/3*dt^3, and that
    // dW and dZ are correlated such that E[dWdZ]=0.5*dt^2. The values are obtained from the following
    //         dW = N1*sqrt(dt)
    //         dZ = 0.5*(N1 + 1/sqrt(3)*N2)*dt^(3/2)
    // where N1 and N2 are independent N(0,1) variables
    m_ppdW = MakeMatrix(m_nWienerSize, m_pDynSystem->GetDimension());
    m_ppdZ = MakeMatrix(m_nWienerSize, m_pDynSystem->GetDimension());

    // Seed generator - either by specified value or time
    if (0 == m_settings.m_nSeed) m_settings.m_nSeed = time(NULL);
    sgenrand(m_settings.m_nSeed);

    float nN1, nN2;
    for (i = 0; i < m_nWienerSize; ++i)
    {
        for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
        {
            RandN(nN1, nN2);

            m_ppdW[i][j] = nN1 * sqrt(m_settings.m_nStepSize);
            m_ppdZ[i][j] = 0.5 * (nN1 + 1/sqrt(3.0) * nN2) * pow(m_settings.m_nStepSize, 3.0/2.0);
        }
    }

    // Should we save this info?
    if (!m_settings.m_strWienerOutputFileName.empty())
    {
        FILE* pFile;
        if (NULL == (pFile = fopen(m_settings.m_strWienerOutputFileName.c_str(), "w")))
        {
            m_strLastError = "Unable to open file for writing Wiener process information: ";
            m_strLastError += m_settings.m_strWienerOutputFileName.c_str();
            return false;
        }

        // Each row corresponds to a time instant. Columns consist dW followed by dZ
        for (i = 0; i < m_nWienerSize; ++i)
        {
            for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
            {
                fprintf(pFile, "%f ", m_ppdW[i][j]);
            }
            for (j = 0; j < m_pDynSystem->GetDimension(); ++j)
            {
                fprintf(pFile, "%f ", m_ppdZ[i][j]);
            }
            fprintf(pFile, "\n");
        }

        fclose(pFile);
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Run
// Class:       CSolver
// Purpose:     Initialises, executes the solver, and writes to file. This is GoF template
//              pattern.
// Inputs:      None
// Outputs:     returns                true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CSolver::Run(void)
{
    bool bRetValue(true);
    if (!Initialise()) return false;

    printf("Solving...\n");
    printf("%s\n", GetDescription().c_str());

    unsigned int d = m_pDynSystem->GetDimension();
    float* pXout = MakeVector(d);
    float* pdW = MakeVector(d);
    float* pdZ = MakeVector(d);
    float ndt = m_settings.m_nStepSize*m_settings.m_nMultiplier;
    float nTime;
    unsigned i, j, k;

    // Create memory to store drift vector and diffusion matrix
    m_pA = MakeVector(d);
    m_ppB = MakeMatrix(d, d);

    // Perform iteration
    for (i = 0; i < m_nSolutionSize - 1; ++i)
    {
        nTime = i*ndt;
        m_ppSolution[i][0] = nTime;    // recall first column of m_ppSolution is time

        m_pDynSystem->GetDriftAndDiffusion(nTime, m_ppSolution[i]+1, m_pA, m_ppB);

        // Prepare correct increment of dW and dZ
        // We do this because the step size of the solver might be a multiple
        // of the Wiener increments. We just sum over the intervening increments
        // as it is really W((i+1)*ndt)-W(i*ndt) that we're after (where ndt is
        // the time step of solver)
        for (j = 0; j < d; ++j)
        {
            pdW[j] = 0;
            pdZ[j] = 0;
            for (k = i*m_settings.m_nMultiplier; k < (i+1)*m_settings.m_nMultiplier; ++k)
            {
                pdW[j] += m_ppdW[k][j];
                pdZ[j] += m_ppdZ[k][j];
            }
        }

        // Call subclass Advance() method where the actual solver is defined
        bRetValue = Advance(nTime, ndt, pdW, pdZ, m_ppSolution[i]+1, m_ppSolution[i+1]+1);

        if (!bRetValue) break;    // solver had a problem?
    }
    m_ppSolution[i][0] = i*ndt;    // final time

    // Write output assuming all went well
    if (bRetValue)
    {
        printf("Writing to file...\n");
        bRetValue = Finish();
        printf("Done.\n");
    }

    // Cleanup
    free(pXout);
    free(pdW);
    free(pdZ);

    return bRetValue;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Finish
// Class:       CSolver
// Purpose:     Write data (solution) to file. This method may be overriden to provide
//              solver-specific cleanup.
// Inputs:      None
// Outputs:     returns             true on success
///////////////////////////////////////////////////////////////////////////////////////////////
bool CSolver::Finish(void)
{
    FILE* pFile = NULL;

    // Write output
    if (NULL == (pFile = fopen((char*)m_settings.m_strOutputFileName.c_str(), "w")))
    {
        m_strLastError = "Unable to open file ";
        m_strLastError += m_settings.m_strOutputFileName;
        return false;
    }

    for (unsigned int i = 0; i < m_nSolutionSize; ++i)
    {
        for (int j = 0; j < m_pDynSystem->GetDimension() + 1; ++j)
        {
            fprintf(pFile, "%f ", m_ppSolution[i][j]);
        }
        fprintf(pFile, "\n");
    }
    fclose(pFile);

    // Perform cleanup
    Cleanup();

    return true;    // success
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    RandN
// Class:       CSolver
// Purpose:     Creates normal random variables with N(0,1)
// Inputs:      None
// Outputs:     rnN1                independent random variable
//              rnN2                independent random variable
///////////////////////////////////////////////////////////////////////////////////////////////
void CSolver::RandN(float& rnN1, float& rnN2)
{
    /*
        Implements the Polar form of the Box-Muller Transformation
        (c) Copyright 1994, Everett F. Carter Jr.

        Permission is granted by the author to use
        this software for any application provided this
        copyright notice is preserved.
    */

    /*
        Uses high quality random number generator MT19937
        Modified to return two independent variables per call
    */

    float nX1, nX2, nW;

    do
    {
        nX1 = 2.0 * genrand() - 1.0;
        nX2 = 2.0 * genrand() - 1.0;
        nW = nX1 * nX1 + nX2 * nX2;
    } while ( nW >= 1.0 );

    nW = sqrt( (-2.0 * log( nW ) ) / nW );

    rnN1 = nX1 * nW;
    rnN2 = nX2 * nW;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    MakeVector
// Class:       CSolver
// Purpose:     Convenience method to make a vector (array of floats initialised to 0 of
//              length i)
// Inputs:      i                   length of vector
// Outputs:     returns             vector
///////////////////////////////////////////////////////////////////////////////////////////////
float* CSolver::MakeVector(size_t i)
{
    return (float*)xcalloc(i, sizeof(float));
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    MakeMatrix
// Class:       CSolver
// Purpose:     Convenience method to make a matrix (multidimensional array of floats
//              initialised to 0 of dimension i x j)
// Inputs:      i                    number of rows
//              j                    number of columns
// Outputs:     returns              matrix
///////////////////////////////////////////////////////////////////////////////////////////////
float** CSolver::MakeMatrix(size_t i, size_t j)
{
    float** pp = (float**)xmalloc(i * sizeof(float*));
    for (unsigned int l = 0; l < i; ++l)
    {
        pp[l] = MakeVector(j);
    }

    return pp;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDescription
// Class:       CSolver
// Purpose:     Get a description of the solver. Subclasses should override this method.
// Inputs:      None
// Outputs:     returns             a short description of the solver
///////////////////////////////////////////////////////////////////////////////////////////////
std::string CSolver::GetDescription(void)
{
    return "Description not available for this solver";
}


