///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    Solver.h
// Package:     SDESolver
// Class:       CSolver
// Purpose:     Base class for Stochastic Differential Equations (SDE) solvers. Uses the
//              Template pattern for defining the solution procedure
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#ifndef SOLVER_H
#define SOLVER_H

#include <string>
#include "DynSystem.h"
#include "SolverSettings.h"

class CSolver
{
    public:

    CSolver(void);
    CSolver(const CSolverSettings& rSettings);
    virtual ~CSolver(void);

    bool Run(void);
    void RegisterDynSystem(CDynSystem* pDynSystem) { m_pDynSystem = pDynSystem; }
    std::string GetLastError(void) { return m_strLastError; }
    float** GetSolution(void) { return m_ppSolution; }

    virtual std::string GetDescription(void);

    protected:

    // Solution procedure - subclasses MUST provide implementations of Advance() and
    // MAY override Initialise() and/or Finish() for finer grain control
    virtual bool Initialise(void);
    virtual bool Advance(float nTime, float ndt, float* pdW, float* pdZ, float* pXin, float* pXout) = 0;
    virtual bool Finish(void);

    // Helper methods
    static float* MakeVector(size_t i);
    static float** MakeMatrix(size_t i, size_t j);
    static void RandN(float& rnN1, float& rnN2);
    void Cleanup(void);

    CSolverSettings m_settings;     // solver independent settings
    CDynSystem* m_pDynSystem;       // system to solve
    std::string m_strLastError;     // last error if any
    float** m_ppdW;                 // dW increments
    float** m_ppdZ;                 // dZ increments (correlated with dW)
    float** m_ppSolution;           // solution
    unsigned int m_nSolutionSize;   // solution size
    unsigned int m_nWienerSize;     // Brownian path size
    float* m_pA;                    // the a coefficient at time t
    float** m_ppB;                  // the b coefficent matrix
};

#endif // SOLVER_H
