///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    SolverFactory.cpp
// Package:     SDESolver
// Class:       CSolverFactory
// Purpose:     Parameterised factory for creating CSolver implementations
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include "SolverFactory.h"
#include "EMSolver.h"
#include "MilsteinSolver.h"
#include "RungeKuttaSolver.h"

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CSolverFactory
// Class:       CSolverFactory
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSolverFactory::CSolverFactory(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    ~CSolverFactory
// Class:       CSolverFactory
// Purpose:     Destructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSolverFactory::~CSolverFactory(void)
{
    // Empty
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CreateSolver
// Class:       CSolverFactory
// Purpose:     Create an engine
// Inputs:      eType               type of solver
//              rSettings           settings for solver
// Outputs:     return              pointer to the new solver or NULL if not a known type
///////////////////////////////////////////////////////////////////////////////////////////////
CSolver* CSolverFactory::CreateSolver(SolverType eType, const CSolverSettings& rSettings)
{
    CSolver* pSolver = NULL;

    switch (eType)
    {
        case CSolverFactory::MILSTEIN:
        {
            pSolver = new CMilsteinSolver(rSettings);
            break;
        }
        case CSolverFactory::RUNGE_KUTTA:
        {
            pSolver = new CRungeKuttaSolver(rSettings);
            break;
        }
        case CSolverFactory::EM:
        {
            pSolver = new CEMSolver(rSettings);
            break;
        }

        // Add implementations here
    }

    return pSolver;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CreateSolver
// Class:       CSolverFactory
// Purpose:     Create an engine
// Inputs:      rstrType            name of solver type
//              rSettings           settings for solver
// Outputs:     return              pointer to the new solver or NULL if not a known type
///////////////////////////////////////////////////////////////////////////////////////////////
CSolver* CSolverFactory::CreateSolver(const std::string& rstrType, const CSolverSettings& rSettings)
{
    return CreateSolver(GetType(rstrType), rSettings);
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetName
// Class:       CSolverFactory
// Purpose:     Get the name of the solver
// Inputs:      eType               type of solver to create
// Outputs:     return              name of solver
///////////////////////////////////////////////////////////////////////////////////////////////
std::string CSolverFactory::GetName(CSolverFactory::SolverType eType)
{
    std::string strName;

    if (CSolverFactory::EM == eType)
    {
        strName = "EM";
    }
    else if (CSolverFactory::MILSTEIN == eType)
    {
        strName = "MILSTEIN";
    }
    else if (CSolverFactory::RUNGE_KUTTA == eType)
    {
        strName = "RUNGE_KUTTA";
    }

    // Add implementations here

    return strName;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetType
// Class:       CSolverFactory
// Purpose:     Get the solver type from name
// Inputs:      rstrName            name of solver
// Outputs:     return              the type
///////////////////////////////////////////////////////////////////////////////////////////////
CSolverFactory::SolverType CSolverFactory::GetType(const std::string& rstrName)
{
    if (rstrName == "EM")
    {
        return CSolverFactory::EM;
    }
    else if (rstrName == "MILSTEIN")
    {
        return CSolverFactory::MILSTEIN;
    }
    else if (rstrName == "RUNGE_KUTTA")
    {
        return CSolverFactory::RUNGE_KUTTA;
    }
    else
    {
        printf("Unrecognised solver name. Using EM\n");
        return CSolverFactory::EM;    // default
    }

    // Add implementations here
}
