///////////////////////////////////////////////////////////////////////////////////////////////
// Filename:    SteinerDynSystem.cpp
// Package:     SDESolver
// Class:       CDynSystem
// Purpose:     SDE system in Ito form:
//                  dX1 = [dH/dX2 - l*(H)*dH/dX1]dt + [-l*sigma*dH/dX1]dW
//                  dX2 = [-dH/dX1 - l*(H)*dH/dX2]dt + [-l*sigma*dH/dX2]dW              ...1
//
//              where W is Wiener process and formally dW/dt = n(t) white noise, and
//              H is the equation for the Steiner Hypocycloid:
//                  (X1^2+X2^2)^2 + 8*r*X1(3*X2^2-X1^2) + 18*r^2*(X1^2+X2^2)-27*r^4     ...2
//              and l > 0, sigma is noise intensity, r = 1
//
// Written:     Dave Burke        03/08/02
///////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "SteinerDynSystem.h"

#define DIM 2    // dimension 2

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    CSteinerDynSystem
// Class:       CSteinerDynSystem
// Purpose:     Default constructor
// Inputs:      None
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
CSteinerDynSystem::CSteinerDynSystem(void)
{
    // Register our dimension
    m_nDimension = DIM;

    // System parameters
    m_nR = 1;
    m_nL = 0.1;
    m_nS = 0.2;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    GetDriftAndDiffusion
// Class:       CSteinerDynSystem
// Purpose:     Get the drift vector (a) and diffusion matrix (b) for the system
// Inputs:      nTime                time
//              pXin                X vector at this time
// Outputs:     pAout                a vector
//              pBout                b matrix
///////////////////////////////////////////////////////////////////////////////////////////////
void CSteinerDynSystem::GetDriftAndDiffusion(float nTime, float* pXin,
                                             float* pAout, float** ppBout)
{
    float X1 = pXin[0];
    float X2 = pXin[1];
    float nH, nH1, nH2;

    nH = pow((pow(X1,2) + pow(X2, 2)),2) + 8*m_nR*X1*(3*pow(X2,2) - pow(X1,2))
       + 18*pow(m_nR,2)*(pow(X1,2) + pow(X2,2)) - 27*pow(m_nR,4);
    nH1 = 2*(pow(X1,2) + pow(X2,2))*2*X1 + 8*m_nR*3*pow(X2,2) - 3*8*m_nR*pow(X1,2) + 18*pow(m_nR,2)*2*X1;
    nH2 = 2*(pow(X1,2) + pow(X2,2))*2*X2 + 8*m_nR*3*X1*2*X2 + 2*18*pow(m_nR,2)*X2;

    // SDE a vector
    pAout[0] = nH2 - m_nL*nH*nH1;
    pAout[1] = -nH1 - m_nL*nH*nH2;

    // SDE b matrix
    ppBout[0][0] = m_nS*m_nL*nH1;
    ppBout[0][1] = 0;
    ppBout[1][0] = 0;
    ppBout[1][1] = m_nS*m_nL*nH2;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Create
// Purpose:     Create an instance of object. Used in shared libraries
// Inputs:      None
// Outputs:     return            pointer to new instance
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" CDynSystem* Create(void)
{
    return new CSteinerDynSystem();
}

///////////////////////////////////////////////////////////////////////////////////////////////
// Function:    Destroy
// Purpose:     Destroy an instance of object. Used in shared libraries
// Inputs:      pDynSystem        object to destroy
// Outputs:     None
///////////////////////////////////////////////////////////////////////////////////////////////
extern "C" void Destroy(CDynSystem* pDynSystem)
{
    delete pDynSystem;
}



