///////////////////////////////////////////////////////////////////////////////
//    Charles River Analytics Inc.
//    Copyright 2002-2006. All Rights Reserved.
//
//    ORIGINAL SYSTEM: The Vision Kit Library
//    SUBSYSTEM:      Image
//    $Workfile: SubWindow.h $
//    $Revision: 1 $
//    Author(s): V.Ablavsky, T.Goodsell, J.Pollak, M.Stevens, D. Stouch
//
//    Description
//    ===========
//
// A sub window is a container that will allow iteration over a
// spatial patch of the image
//
///////////////////////////////////////////////////////////////////////////////

#ifndef SubWindow_h
#define SubWindow_h

#include "ImageDefs.h"
#include <iterator>

namespace cuj
{


  //----------------------------------------------------------------------
  template<size_t _SubWidth, size_t _SubHeight, typename _Tp>
    class SubWindow: public std::iterator<heavy_forward_iterator_tag, _Tp>
  {
    
    // simplifying typedef
    typedef SubWindow<_SubWidth, _SubHeight, _Tp> self;
    
    public:

    // forward declaratin
    struct proxy_iterator;
    
    // the iterator
    class iterator: public std::iterator<std::forward_iterator_tag, _Tp>
    {
      friend class SubWindow<_SubWidth, _SubHeight, _Tp>;
      public:
            
      // dereference
      inline _Tp operator*(void) const
      {        
        if ((x_ < 0) || (y_ < 0) || (x_ >= width_) || (y_ >= yOffImage_))
        {
          return 0;
        }
        return begin_[y_ + x_];
      }
      
      // advance using pre-increment
      inline iterator & operator ++(void)
      {
        x_ += 1;
        if (x_ > stopX_)
        {
          x_  = startX_;
          y_ += width_;
        }
        return *this;
      }
      
      // test iteration completion
      inline bool operator!=(const iterator & rhs) const
      {
        return ((x_ != rhs.x_) || (y_ != rhs.y_));
      }
      

      proxy_iterator operator&(void)
      {
        return proxy_iterator(this);
      }
          
      protected:
      
      private:
      
      // constructor        
      iterator(_Tp *in, int width, int height) :
        x_(0),
        y_(0),
        begin_(in),
        width_(width),
        height_(height),
        yOffImage_(height * width)
      {
      }
      
      // current position in the image
      int x_;
      int y_;
      
      // left and right boundaries on the scan line
      int startX_;
      int stopX_;
      
      // the start of the entire image buffer, not the current
      // position defined by x and y
      _Tp *begin_;
      
      // for easy access, store the image position
      int width_;
      int height_;
     int yOffImage_;
      
    }; // end of iterator definition

    struct proxy_iterator: public std::iterator<std::forward_iterator_tag, _Tp>
    {
      typedef proxy_iterator self;

      inline proxy_iterator(iterator* p)
        : addressee_(p)
      {
      }

      inline self& operator++(void)
      {
        addressee_->operator++();
        return *this;
      }

      inline _Tp operator*(void)
      {
        return addressee_->operator*();
      }

      inline bool operator!=(const self & rhs) const
      {
        return (*addressee_ != *rhs.addressee_);
      }


      iterator* addressee_;
    }; // end of proxy_iterator definition
    
    // access 
    inline iterator &begin()
    {
      return begin_;
    }
    inline iterator &end()
    {
      return end_;
    }
        
    // construct container based on image
    SubWindow(_Tp *in, int width, int height) :
      begin_(in, width, height),
      end_  (in, width, height)
    {
    }
    
    // set the position in the larger image
    inline void updateCenter(size_t x, size_t y)
    {
      begin_.startX_  = x - _SubWidth / 2;
      begin_.stopX_   = x + _SubWidth / 2;
      begin_.x_       = begin_.startX_;
      begin_.y_       = (y - _SubHeight/2) * begin_.width_;
      
      end_.x_         = begin_.startX_;
      end_.y_         = (y + _SubHeight/2 + 1) * begin_.width_;
    }
    
    protected:
    
    // current location
    iterator begin_;
    iterator end_;
    
    private:
    
  };
  //----------------------------------------------------------------------
  
}

#endif // SubWindow_h
