
//    Date : Sun Sep  1 20:41:18 EDT 2002
//    Author : Sony E Antony


//    File : ELF_LogManager.h
//    File belongs to ELF : Exception Handling and Logging Framework.
//    sonyeantony@yahoo.com


//////////////////////////////////////////////////////////////////////////
//
//
// Macro s used
// SINGLE_THREAD_PROCESS : Thread Id will not be generated with log messages. Once this is defined,
// there is no need to link with -lpthread library.
// 
//
// NO_CRITICAL, NO_ERROR, NO_WARNING, NO_AUDIT, NO_DEBUG
// Once once of these macros are defined at compile time, the corresponding
// LOG macros ( DEBUG_LOG(), ERROR_LOG() etc. ) will become nil. i. e. No code is generated 
// to actually loag anything
// Also these macros form a hierarcy. If NO_CRITICAL is defined it also defines the ones 
// below it thereby making all log statements void.
//
//
// NO_FUNC_TRACE : Function tracing capability is completely gone.
//
// NO_VERBOSE_THROW : Java like exception' s with stacktrace are gone
//
// NOTE: If both NO_FUNC_TRACE & NO_VERBOSE_THROW are defined at compile time, ELF_FUNC_DEF() and 
// ELF_FUNC_DEF_END expans to be exactly the method signature.
// i. e no extra code will be produced
//
//
//
//////////////////////////////////////////////////////////////////////////

#ifndef __ELF_LOG_MANAGER_H__
#define __ELF_LOG_MANAGER_H__

#include<string>
#include<map>

#include"ELF_Exception.h"
#include"ELF_LogWriter.h"
#include"ELF_Properties.h"

using namespace std ;

#ifdef  NO_CRITICAL
#    define NO_ERROR
#    define NO_WARNING
#    define NO_AUDIT
#    define NO_DEBUG
#endif

#ifdef  NO_ERROR
#    define NO_WARNING
#    define NO_AUDIT
#    define NO_DEBUG
#endif

#ifdef  NO_ERRO
#    define NO_AUDIT
#    define NO_DEBUG
#endif

#ifdef  NO_AUDIT
#    define NO_DEBUG
#endif

#ifdef  NO_FUNC_TRACE
#   ifdef NO_VERBOSE_THROW
#       define ELF_FUNCT_DEF( signature ) signature
#       define ELF_FUNCT_DEF_END
#   else
#       define ELF_FUNCT_DEF( signature ) signature { \
            string ELF_FUNC_NAME( #signature ) ; \
            try

#       define ELF_FUNCT_DEF_END catch( ELF::Exception &e ) { e.addStackFrame( ELF_FUNC_NAME ) ; throw ; } }
#    endif
#else
#   ifdef NO_VERBOSE_THROW
#       define ELF_FUNCT_DEF( signature ) signature { \
            string ELF_FUNC_NAME( #signature ) ;\
            LogManager::FunctCallMonitor tmp_5763( logManager.funcCallTracingIsOn(), ELF_FUNC_NAME ) ;
#       define ELF_FUNCT_DEF_END }
#   else
#       define ELF_FUNCT_DEF( signature ) signature { \
            string ELF_FUNC_NAME( #signature ) ;\
            LogManager::FunctCallMonitor tmp_5763( logManager.funcCallTracingIsOn(), ELF_FUNC_NAME ) ; \
            try
#       define ELF_FUNCT_DEF_END catch( ELF::Exception &e ) { e.addStackFrame( ELF_FUNC_NAME ) ; throw ; } }
#   endif
#endif

#ifdef  NO_CRITICAL
#    define CRIT_LOG( dataString )
#else
#    define CRIT_LOG( dataString ) logManager.logCritical( \
        string( "--CRITICAL MSG BEGIN--" ) + \
        ELF_FUNC_NAME + "\n" + string( dataString ) + \
        string( "\n--CRITICAL MSG END--") )
#endif

#ifdef  NO_ERROR
#    define ERROR_LOG( dataString )
#else
#    define ERROR_LOG( dataString ) logManager.logError( \
        string( "--ERROR MSG BEGIN--" ) + \
        ELF_FUNC_NAME + "\n" + string( dataString ) + \
        string( "\n--ERROR MSG END--") )
#endif

#ifdef  NO_WARNING
#    define WARNG_LOG( dataString )
#else
#    define WARNG_LOG( dataString ) logManager.logWarning( \
        string( "--WARNING MSG BEGIN--" ) + \
        ELF_FUNC_NAME + "\n" + string( dataString ) + \
        string( "\n--WARNING MSG END--") )
#endif

#ifdef  NO_AUDIT
#    define AUDIT_LOG( dataString )
#else
#    define AUDIT_LOG( dataString ) logManager.logAudit( \
        string( "--AUDIT MSG BEGIN--" ) + \
        ELF_FUNC_NAME + "\n" + string( dataString ) + \
        string( "\n--AUDIT MSG END--") )
#endif

#ifdef  NO_DEBUG
#    define DEBUG_LOG( dataString )
#else
#    define DEBUG_LOG( dataString ) logManager.logDebug( \
        string( "--DEBUG MSG BEGIN--" ) + \
        ELF_FUNC_NAME + "\n" + string( dataString ) + \
        string( "\n--DEBUG MSG END--") )
#endif

namespace ELF {

DEFINE_EXCEPTION( LogManagerException ) ;

class LogManager {

    typedef enum { LOG_NONE = 0, LOG_CRITICAL = 1, LOG_ERROR = 2, LOG_WARNING = 3, LOG_AUDIT = 4,
            LOG_DEBUG = 5 } DataLogLevel ;

    Properties props_ ;
    DataLogLevel dataLogLevel_ ;
    bool initialized_ ;
    bool funcCallTracingIsOn_ ;
    void (LogWriter::*logRoutinePtrs_[6])( const string& ) ;
    string pid_ ;
    string logFile_ ;
    LogWriter *logWriter_ ;

  public:

    class FunctCallMonitor {

        const string &funSig_ ;
        bool &trcngIsOn_ ;

      public:

        FunctCallMonitor( bool &trcngIsOn,  const string &sig ) ;
        ~FunctCallMonitor( ) ;

    };

    typedef enum { FILEBASED = 0, UDPBASED = 1, TCPBASED = 2 } LoggingType ;

    LogManager() ;
    ~LogManager() ;

    void initialize( LoggingType type, PropertyList props = appConfig.getPropertyList() ) throw( Exception ) ;


    inline void logCritical( const string& msg ) throw() {
        logMessage_( msg, 0 ) ;
        }

    inline void logError( const string& msg ) throw() {
        logMessage_( msg, 1 ) ;
        }

    inline void logWarning( const string& msg ) throw() {
        logMessage_( msg, 2 ) ;
        }

    inline void logAudit( const string& msg ) throw() {
        logMessage_( msg, 3 ) ;
        }

    inline void logDebug( const string& msg ) throw() {
        logMessage_( msg, 4 ) ;
        }

    inline void logFunc( const string& msg ) throw() {
        logMessage_( msg, 5 ) ;
        }

    void turnOnFuncCallTracing( ) { funcCallTracingIsOn_ = true ; logRoutinePtrs_[5] = &LogWriter::logFunctCall ; }
    void turnOffFuncCallTracing( ) { funcCallTracingIsOn_ = false ; logRoutinePtrs_[5] = &LogWriter::dummy ; }
    inline bool& funcCallTracingIsOn( )  { return funcCallTracingIsOn_ ; }
    void addLogWriter( LoggingType type, Properties props = appConfig.getPropertyList() ) throw( Exception ) ;
    bool removeTailEndChainedWriter( ) { return logWriter_->removeTailEndChainedWriter() ; }

    void setDataLogLevel( DataLogLevel l )  ;
    inline DataLogLevel getDataLogLevel( )  { return dataLogLevel_ ; }


  private :

    void writeLogManagerMsgs_( const string  &msg ) throw() ;
    void logMessage_( const string& msg, int level ) throw( ) ;

  };

    extern LogManager logManager ;
}

using namespace ELF ;
#endif
