
//    Date : Sun Sep  1 20:09:28 EDT 2002
//    Author : Sony E Antony


//    File : ELF_UdpLogServerImpl.C
//    File belongs to ELF : Exception Handling and Logging Framework.
//    sonyeantony@yahoo.com

#include<sys/types.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<unistd.h>
#include<strings.h>
#include<errno.h>

#include "ELF_ConnectionTypes.h"
#include "ELF_UdpLogServerImpl.h"
#include "ELF_MiscUtils.h"


using namespace ELF ;

using namespace std ;

UdpLogServerImpl::UdpLogServerImpl( ) throw ( Exception ) {

    string port = appConfig.getProperty( "ELF_LOG_SERVER_UDP_PORT" ) ;
    const char* portstr = port.c_str() ;
    char* cp ;
    port_ = strtol( portstr, &cp, 10 ) ;
    if( ( *cp != '\0' ) || ( port_ <= 0 ) || ( port_ > 65536 ) ) {
        ostringstream os ;
        os<<"Failed to convert the port to a valid UDP port ( 0 < validport < 65536 ) : "<<port<<endl ;
        os<<strerror( errno ) ;
        THROW_EXCEPTN_WWHERE( UdpLogServerImplException, \
            "UdpLogServerImpl::UdpLogServerImpl( )", os.str() ) ;
    }
}

UdpLogServerImpl::~UdpLogServerImpl( ) {
    close( socket_ ) ;
}

void UdpLogServerImpl::setupSocket_() throw ( Exception ) {

    struct sockaddr_in servsock ;
    if( -1 == ( socket_ = socket( AF_INET, SOCK_DGRAM, 0 ) ) ) {
        ostringstream os ;
        os<<"Failed to create TCPv4 STREAM socket : "<<endl ;
        os<<strerror( errno ) ;
        THROW_EXCEPTN_WWHERE( UdpLogServerImplException, "void UdpLogServerImpl::setupSocket_()", os.str() ) ;
    }
    FileGuard guard( socket_ ) ;
    bzero( &servsock, sizeof( servsock ) ) ;
    servsock.sin_family = AF_INET ;
    servsock.sin_addr.s_addr = htonl( INADDR_ANY ) ;
    servsock.sin_port = htons( port_ ) ;
    if( -1 == bind( socket_, (struct sockaddr*) &servsock, sizeof( servsock ) ) ) {
        ostringstream os ;
        os<<"Failed to bind() UDP socket to port : "<< port_<<endl ;
        os<<strerror( errno ) ;
        THROW_EXCEPTN_WWHERE( UdpLogServerImplException, "void UdpLogServerImpl::setupSocket_()", os.str() ) ;
    }

    int on = 1 ;
    socklen_t len = sizeof( on ) ;
    if( -1 == setsockopt( socket_, SOL_SOCKET, SO_REUSEADDR, &on, len ) ) {
        ostringstream os ;
        os<<"Failed to set REUSEADDR socket option for socket at port : "<< port_<<endl ;
        os<<strerror( errno ) ;
        THROW_EXCEPTN_WWHERE( UdpLogServerImplException, "void UdpLogServerImpl::setupSocket_()", os.str() ) ;
    }
    guard.noClose() ;
}

void UdpLogServerImpl::run( BlockingOperationSentry* callbackObj ) throw ( Exception ) {

    callbackObj_ = callbackObj ;
    setupSocket_() ;
    string clientIp ;
    while( 1 ) {

        try {
            readMessage_( clientIp ) ;
            processMessage_( clientIp ) ;
        }
        catch( Exception &e ) {
            writeLogServerMsgs( e.toString() ) ;
        }
        catch( ... ) {
            writeLogServerMsgs( "Unknown Exception Caught in UdpLogServerImpl::run()" ) ;
        }
    }
}


void UdpLogServerImpl::readMessage_( string &clientIp ) throw ( Exception ) {

    struct sockaddr_in cliaddr ;
    socklen_t len = sizeof( cliaddr ) ;
    int msglen = 0 ;
    {
        BlockingOperationSentryGuard guard( callbackObj_ ) ;
        errno = 0 ;
        if( -1 == ( msglen = recvfrom( socket_, &msg_, sizeof( msg_) - 1, 0,
                (struct sockaddr*)&cliaddr, &len ) ) ) {

            ostringstream os ;
            os<<"Failed to read from socket : "<<endl ;
            os<<strerror( errno ) ;
            THROW_EXCEPTN_WWHERE( UdpLogServerImplException,\
                    "void UdpLogServerImpl::readMessage_( string &clientIp )", os.str() ) ;
        }
    }
    msg_.buf[msglen] = '\0' ;
    char buf[50] ;
    if( inet_ntop( AF_INET, &cliaddr.sin_addr, buf, sizeof( buf ) ) == NULL ) {
        clientIp = "" ;
    }
    else {
        clientIp = buf + string( " : " ) ;
    }
}

void UdpLogServerImpl::processMessage_( const string &clientIp ) throw ( Exception ) {

    if( ConnectionTypes::CRITICAL == msg_.body.type ) {
        logWriter_->logCritical( clientIp + msg_.body.text ) ;
    }
    else if( ConnectionTypes::ERROR == msg_.body.type ) {
        logWriter_->logError( clientIp + msg_.body.text ) ;
    }
    else if( ConnectionTypes::WARNING == msg_.body.type ) {
        logWriter_->logWarning( clientIp + msg_.body.text ) ;
    }
    else if( ConnectionTypes::AUDIT == msg_.body.type ) {
        logWriter_->logAudit( clientIp + msg_.body.text ) ;
    }
    else if(  ConnectionTypes::DEBUG == msg_.body.type ) {
        logWriter_->logDebug( clientIp + msg_.body.text ) ;
    }
    else if( ConnectionTypes::FUNCALL == msg_.body.type ) {
        logWriter_->logFunctCall( clientIp + msg_.body.text ) ;
    }
    else {
        ostringstream os ;
        os<<"Unknown type of message received FROM : "<<clientIp<<". TYPE : "<<msg_.body.type<<endl ;
        os<<"TEXT : "<<msg_.body.text ;
        THROW_EXCEPTN_WWHERE( UdpLogServerImplException,\
            "void UdpLogServerImpl::processMessage_( const string &clientIp )", os.str() ) ;
    }
}

