//+---------------------------------------------------------------------------
//
//  Microsoft Windows
//  Copyright (C) Microsoft Corporation, 2002.
//
//  File:       foreach.h
//
//  Contents:   FOR_EACH macro and supporting templates
//
//  Classes:    is_managed, auto_unbox, enumerator<>
//
//  Functions:
//
//  Notes:
//
//  History:    11-18-2002   ansont     Created
//              11-19-2002   ericne     ported to VC7
//              11-20-2002   ericne     eliminate gcroot for better perf
//              11-21-2002   ericne     rename FOREACH to FOR_EACH
//              11-24-2002   ericne     ArrayList and BitArray use indexing
//                                      instead of enumeration
//              11-25-2002   ericne     hoist length calculation out of loop
//              11-27-2002   ericne     Iterating over 2+ dim managed array 
//                                      is now a compile error
//              01-12-2003   ericne     Support iterating over STL containers
//              01-16-2003   ericne     Support iterating over native strings
//                                      Disallow UDCs in enumerator_selector
//              01-21-2003   ericne     Support System::Text::StringBuilder
//              09-22-2003   ericne     Support by-ref iteration
//
//----------------------------------------------------------------------------

#ifdef _MSC_VER
# pragma once
#endif

#ifndef FOR_EACH

#if defined(_MSC_VER) & defined(_MANAGED)
# using <mscorlib.dll>
#endif

#include <cstddef>

#ifdef _MSC_VER
# include <crtdbg.h> // prefer _ASSERTE for better diagnostics on Win32
# define ASSERT_FE(expr) _ASSERTE(expr)
# define DUMMY_FE ,0,0
# define DEFAULT_FE ,int
#else
#include <cassert>  // default assert for non-Win32
# define ASSERT_FE(expr) assert(expr)
# define __cdecl
# define DUMMY_FE
# define DEFAULT_FE =0
#endif

namespace for_each
{

// wrap_type is used to disallow user-defined conversions from happening
// to parameters to the enum_select functions.
template<typename T>
struct wrap_type
{
    wrap_type( T );
};

typedef char (&unknown_type       )[1];
typedef char (&IEnumerable_type   )[2];
typedef char (&String_type        )[3];
typedef char (&StringBuilder_type )[4];
typedef char (&Array_type         )[5];
typedef char (&ArrayList_type     )[6];
typedef char (&BitArray_type      )[7];
typedef char (&native_string_type )[8];
typedef char (&native_array_type  )[9];
typedef char (&std_container_type )[10];

#if defined(_MSC_VER) & defined(_MANAGED)
Array_type           enum_select( System::Array*,                     wrap_type<System::Array*>,                       int DEFAULT_FE );
String_type          enum_select( System::String*,                    wrap_type<System::String*>,                      int DEFAULT_FE );
StringBuilder_type   enum_select( System::Text::StringBuilder*,       wrap_type<System::Text::StringBuilder*>,         int DEFAULT_FE );
BitArray_type        enum_select( System::Collections::BitArray*,     wrap_type<System::Collections::BitArray*>,       int DEFAULT_FE );
ArrayList_type       enum_select( System::Collections::ArrayList*,    wrap_type<System::Collections::ArrayList*>,      int DEFAULT_FE );
IEnumerable_type     enum_select( System::Collections::IEnumerable*,  wrap_type<System::Collections::IEnumerable*>,    int DEFAULT_FE );
#endif

native_string_type   enum_select( char const*,    wrap_type<char const *>,     int DEFAULT_FE );
native_string_type   enum_select( wchar_t const*, wrap_type<wchar_t const*>,   int DEFAULT_FE );
template<typename T,size_t N>
native_array_type    enum_select( T (&)[N], T (&)[N], int DEFAULT_FE );
template<typename T>
std_container_type   enum_select( T &, T &, typename T::const_iterator * DEFAULT_FE );
unknown_type __cdecl enum_select( ... );

template<size_t>
struct enumerator;

// leave this undefined
struct do_not_know_how_to_loop_over_type;

template<>
struct enumerator<sizeof(unknown_type)>
{
    typedef do_not_know_how_to_loop_over_type enm_t;
    typedef int len_t;
    static bool __cdecl nul_enm( ... );
    static bool __cdecl nul_len( ... );
    static bool __cdecl init( ... );
    static bool __cdecl next( ... );
};

template<typename T> struct factory
{
    static T get();
};

#if defined(_MSC_VER) & defined(_MANAGED)

char             check_is_managed( wrap_type<System::Object *> );
double __cdecl   check_is_managed( ... );

template<typename T>
struct is_managed
{
    enum { value = 1==sizeof(check_is_managed(factory<T>::get())) };
};

template<bool>
struct auto_unbox_helper;

template<>
struct auto_unbox_helper<true> // managed
{
    template<typename T>
    struct inner
    {
        static T unbox( System::Object *o )
        {
            return __try_cast<T>( o );
        }
    };
};

template<>
struct auto_unbox_helper<false> // not managed
{
    template<typename T>
    struct inner
    {
        static T __gc& unbox( System::Object *o )
        {
            return *__try_cast<T __box*>( o );
        }
    };
};

template<typename T>
struct auto_unbox
    : auto_unbox_helper<is_managed<T>::value>::template inner<T>
{
};

template<>
struct enumerator<sizeof(IEnumerable_type)>
{
    typedef System::Collections::IEnumerator* enm_t;
    typedef int len_t;

    static enm_t nul_enm( System::Collections::IEnumerable * )
    {
        return 0;
    }
    static len_t nul_len( System::Collections::IEnumerable * )
    {
        return 0;
    }
    static bool init( enm_t &e, len_t, System::Collections::IEnumerable *c )
    {
        e = c->GetEnumerator();
        return false;
    }
    template<typename T>
    static bool next( T &var, System::Collections::IEnumerable *, enm_t &e, len_t )
    {
        if( ! e->MoveNext() )
            return false;

        var = auto_unbox<T>::unbox( e->Current );
        return true;
    }
};

template<typename StringType>
struct string_enumerator
{
    typedef int enm_t;
    typedef int len_t;

    static enm_t nul_enm( StringType * )
    {
        return 0;
    }
    static len_t nul_len( StringType * )
    {
        return 0;
    }
    static bool init( enm_t, len_t &len, StringType *str )
    {
        len = str->Length;
        return false;
    }
    template<typename T>
    static bool next( T &var, StringType *str, enm_t &ind, len_t len )
    {
        // if this assert fails, the string has been modified during iteration
        ASSERT_FE( len == str->Length );
        if( ind >= len )
            return false;
        
        var = str->Chars[ind];
        ++ind;
        return true;
    }                  
};

template<>
struct enumerator<sizeof(String_type)> : string_enumerator<System::String>
{
};

template<>
struct enumerator<sizeof(StringBuilder_type)> : string_enumerator<System::Text::StringBuilder>
{
};

template<>
struct enumerator<sizeof(Array_type)>
{
    typedef int enm_t;
    typedef int len_t;

    static enm_t nul_enm( System::Array * )
    {
        return 0;
    }
    static len_t nul_len( System::Array * )
    {
        return 0;
    }
    static bool init( enm_t, len_t &len, System::Array *arr )
    {
        ASSERT_FE( arr->Rank == 1 ); // can only iterate 1-D arrays
        len = arr->Length;
        return false;
    }
    template<typename T,typename U>
    static bool next( T &var, U &u, enm_t &ind, len_t len )
    {
        return move_next_impl( var, u, ind, len, &u );
    }
private:
    template<typename T>
    static bool move_next_impl( T &var, System::Array *arr, enm_t &ind, len_t len, System::Array ** )
    {
        // if this assert fails, the array has been modified during iteration
        ASSERT_FE( len == arr->Length );
        if( ind >= len )
            return false;
        
        var = auto_unbox<T>::unbox( arr->GetValue(ind) );
        ++ind;
        return true;
    }
    template<typename T,typename U>
    static bool move_next_impl( T &var, U (&arr) __gc[], enm_t &ind, len_t len, U (*) __gc[] )
    {
        // if this assert fails, the array has been modified during iteration
        ASSERT_FE( len == arr->Length );
        if( ind >= len )
            return false;
        
        var = arr[ind];
        ++ind;
        return true;
    }
};

template<>
struct enumerator<sizeof(ArrayList_type)>
{
    typedef int enm_t;
    typedef int len_t;

    static enm_t nul_enm( System::Collections::ArrayList * )
    {
        return 0;
    }
    static len_t nul_len( System::Collections::ArrayList * )
    {
        return 0;
    }
    static bool init( enm_t, len_t &len, System::Collections::ArrayList *arr )
    {
        len = arr->Count;
        return false;
    }
    template<typename T>
    static bool next( T &var, System::Collections::ArrayList *arr, enm_t &ind, len_t len )
    {
        // if this assert fails, the ArrayList has been modified during iteration
        ASSERT_FE( len == arr->Count );
        if( ind >= len )
            return false;
        
        var = auto_unbox<T>::unbox( arr->Item[ind] );
        ++ind;
        return true;
    }
};

template<>
struct enumerator<sizeof(BitArray_type)>
{
    typedef int enm_t;
    typedef int len_t;

    static enm_t nul_enm( System::Collections::BitArray * )
    {
        return 0;
    }
    static len_t nul_len( System::Collections::BitArray * )
    {
        return 0;
    }
    static bool init( enm_t, len_t &len, System::Collections::BitArray *arr )
    {
        len = arr->Length;
        return false;
    }
    template<typename T>
    static bool next( T &var, System::Collections::BitArray *arr, enm_t &ind, len_t len )
    {
        // if this assert fails, the BitArray has been modified during iteration
        ASSERT_FE( len == arr->Length );
        if( ind >= len )
            return false;
        
        var = arr->Item[ind];
        ++ind;
        return true;
    }
};

#endif // _MANAGED

struct ref_base
{
};

template<typename T>
struct ref : ref_base
{
    typedef T type;

    T * m_ptr;
    
    operator T&() const
    {
        return *m_ptr;
    }
};

char            check_ref_wrap( ref_base const & );
double __cdecl  check_ref_wrap( ... );

template<typename T>
struct is_ref_wrap
{
    enum { value = 1==sizeof(check_ref_wrap(factory<T>::get())) };
};

template<bool>
struct add_ref_helper
{
    template<typename T>
    struct inner
    {
        typedef T & type;
    };
};

template<>
struct add_ref_helper<true>
{
    template<typename T>
    struct inner
    {
        typedef typename T::type & type;
    };
};

template<typename T>
struct add_ref
    : add_ref_helper<is_ref_wrap<T>::value>::template inner<T>
{
};

template<bool>
struct assign_helper
{
    template<typename T, typename U>
    static void assign( T & t, U & u )
    {
        t = u;
    }
};

template<>
struct assign_helper<true>
{
    template<typename T,typename U>
    static void assign( ref<T> & t, U & u )
    {
        t.m_ptr = &u;
    }
};

template<typename T>
struct auto_assign 
    : assign_helper<is_ref_wrap<T>::value> 
{
};

// for looping over null-terminated strings
template<>
struct enumerator<sizeof(native_string_type)>
{
    typedef int enm_t;
    typedef int len_t;

    template<typename CH>
    static enm_t nul_enm( CH const * )
    {
        return 0;
    }
    template<typename CH>
    static len_t nul_len( CH const * )
    {
        return 0;
    }
    template<typename CH>
    static bool init( enm_t, len_t, CH const * )
    {
        return false;
    }
    template<typename T,typename CH>
    static bool next( T &var, CH *sz, enm_t &ind, len_t )
    {
        if( ! sz[ind] )
            return false;
        
        auto_assign<T>::assign( var, sz[ind] );
        ++ind;
        return true;
    }
};

// for looping over native arrays
template<>
struct enumerator<sizeof(native_array_type)>
{
    typedef int enm_t;
    typedef int len_t;

    template<typename T,size_t N>
    static enm_t nul_enm( T (&)[N] )
    {
        return 0;
    }
    template<typename T,size_t N>
    static len_t nul_len( T (&)[N] )
    {
        return 0;
    }
    template<typename T,size_t N>
    static bool init( enm_t, len_t, T (&)[N] )
    {
        return false;
    }
    template<typename T,typename U,size_t N>
    static bool next( T &var, U (&arr)[N], enm_t &ind, len_t )
    {
        if( ind >= N )
            return false;
        
        auto_assign<T>::assign( var, arr[ind] );
        ++ind;
        return true;
    }
};

char    check_is_const( const volatile void * );
double  check_is_const( volatile void * );

template<typename T>
struct is_const
{
    enum { value = 1==sizeof(check_is_const(static_cast<T*>(0))) };
};

template<bool>
struct iter_helper
{
    template<typename C> struct inner
    {
        typedef typename C::iterator type;
    };
};

template<>
struct iter_helper<true>
{
    template<typename C> struct inner
    {
        typedef typename C::const_iterator type;
    };
};

template<typename C>
struct iter
    : iter_helper<is_const<C>::value>::template inner<C>
{
};

// generic_type/concrete_type is a poor man's polymorphism.
// It gives fast and cheap polymorphism without virtuals.
struct generic_type
{
    operator bool() const
    {
        return false;
    }
};

template<typename T>
struct concrete_type : generic_type
{
    concrete_type( T const & item )
        : m_item( item )
    {
    }
    
    mutable T m_item;
};

template<typename T>
inline T & generic_cast( generic_type const & g )
{
    return static_cast<concrete_type<T> const &>( g ).m_item;
}

// for looping over anything that looks like an STL container
// (must have const_iterator nested typedef, begin() and end()
// member functions.)
template<>
struct enumerator<sizeof(std_container_type)>
{
    typedef generic_type const & enm_t;
    typedef generic_type const & len_t;

    template<typename C>
        static concrete_type<typename iter<C>::type> nul_enm( C &coll )
    {
        return coll.begin();
    }
    template<typename C>
    static concrete_type<typename iter<C>::type> nul_len( C &coll )
    {
        return coll.end();
    }
    template<typename C>
    static bool init( enm_t, len_t, C const & )
    {
        return false;
    }
    template<typename T,typename C>
    static bool next( T &var, C &coll, enm_t icur, len_t iend )
    {
        typedef typename iter<C>::type iter_type;

        // if this assert fails, the collection has been modified during iteration
        ASSERT_FE( coll.end() == generic_cast<iter_type>( iend ) );
        iter_type &it = generic_cast<iter_type>( icur );
        if( it == generic_cast<iter_type>( iend ) )
            return false;

        auto_assign<T>::assign( var, *it );
        ++it;
        return true;
    }
};

template<typename T>
struct ref_traits
{
    typedef T type;
};

#if !defined(_MSC_VER) | _MSC_VER >= 1310

template<typename T>
struct ref_traits<T&>
{
    typedef ref<T> type;
};

#endif

} // namespace for_each

#define ENM_FE( col )                                                                     \
    for_each::enumerator< sizeof( for_each::enum_select( col, col DUMMY_FE ) ) >

#define VAL_FE( typ )                                                                     \
    for_each::ref_traits< typ >::type

#define REF_FE( typ )                                                                     \
    for_each::add_ref< VAL_FE( typ ) >::type

#define FOR_EACH( typ, var, col )                                                         \
    if( bool _done = false )                                                            {}\
    else if( ENM_FE( col )::enm_t _enm = ENM_FE( col )::nul_enm( col ) )                {}\
    else if( ENM_FE( col )::len_t _len = ENM_FE( col )::nul_len( col ) )                {}\
    else if( ENM_FE( col )::init( _enm, _len, col ) )                                   {}\
    else for( VAL_FE( typ ) _tmp; !_done && ENM_FE( col )::next( _tmp, col, _enm, _len ); )\
        if( !( _done = true ) )                                                         {}\
        else for( REF_FE( typ ) var = _tmp; _done; _done = false )

#endif // #ifndef FOR_EACH

