// Integrator class definitions

// This file contains the class templates for the integrator
// package only.  The implementation of this can be found 
// in a file called integrators.h. 
// This header has been split into two files to allow for
// pre-compilation of frequently used instances.
// Programs which DO NOT use precompiled instances should
//	include integrators.h, which will automatically
//	include this file.

// Currently available: Euler, Runge-Kutta and Gear.

// All integrators have been successfully compiled with the 
// following compilers:

// Sun Solaris Workshop C++ 5.0
// gcc 2.95.2
 

#ifndef INTEGRATOR_DEF_H
#define INTEGRATOR_DEF_H

#include <deque>
#include <stdexcept>


template <class C, class T> class Integrator
{
    protected:
        C statevector;
        T t, h;
        void (*return_derivs)(const C&, const T, C&);

        Integrator (const C&, const T, const T,
            void (*)(const C&, const T, C&));

    public:
        // Virtual destructor allows deletion of derived objects
        virtual ~Integrator() { }

        // Every integrator must have a step() function which
        // updates the state.
        virtual void step(void) = 0;

        // Function to retrieve the current time
        T current_time(void) const {return(t);}

        // Function to retrieve the current state of the system
        void get_state(C&, T&) const;

        // Function to (re)set h
        virtual void set_step_size(T new_h) {h = new_h;}
};


// Euler integrator
template <class C, class T, class B> class Euler
    : public Integrator<C,T>
{
    private:
        C dxdt;

    public:

        Euler (const C &x0, const T t0, const T h_init,
            void (*dxdt_calc)(const C&, const T, C&))
            :Integrator <C,T> (x0, t0, h_init, dxdt_calc),
		    dxdt (x0) { }

        void step(void);
};

// Gear integrator
template <class C, class T, class B> class Gear
    : public Integrator<C,T>
{
    private:
        // A function which compares two objects of class C
        // for approximate equality.
        bool (*approx_equal)(const C&, const C&);
    
        // Order of method and maximum number of iterations
        const unsigned int order, maxiter;
        unsigned int curr_order;
   
        // Storage for past values.
        std::deque<C> past_state;

        // Temporary work spaces
        C xnext, xtemp, dxdt;

        // The virtual function step() evaluates the 
        // Gear formulas through an array of pointers.
        typedef void (Gear<C,T,B>::*xnext_functions)(void);
        xnext_functions calc_xnext[4];

        // The following functions evaluate the Gear formulas
        // of specified order, loading the results into xnext.
        void Gear1xnext(void);
        void Gear2xnext(void);
        void Gear3xnext(void);
        void Gear4xnext(void);

    public:
        Gear(const unsigned int, const C &,
             const T, const T, 
             const unsigned int,
             void (*)(const C&, const T, C&),
             bool (*)(const C&, const C&));
        void step(void);
        void set_step_size(T);
};

// 4th order Runge-Kutta method
template <class C, class T, class B> class RK4
            : public Integrator <C,T>

{
    private:
        C k1;
        C k2;
        C k3;
        C k4;

    public:
      RK4(const C &x0, const T t0, const T h_init,
          void (*dxdt_calc)(const C& , const T, C& ))
         :Integrator <C,T> (x0, t0, h_init, dxdt_calc),
          k1 (x0), k2 (x0), k3 (x0), k4 (x0) { }
      void step (void);
};

#endif // INTEGRATOR_DEF_H
