// Integrator classes


// This file contains only the implementation of the integrator
// package. The class definitions are contained in  a file 
// called integrators_def.h.
// The complete integrator header file has been split into two
// files, integrator_def.h which contains the class definitions,
// and this file, integrators.h, which contains the class 
// implementations.  This splitting was done to allow for
// the pre-compilation of common instances.
// Including this file automatically includes integrator_def.h.


// This code has been successfully compiled with the 
// following compilers:

// Sun Solaris Workshop C++ 5.0
// gcc version 2.95.2

// The integrators currently available in this file include
// Euler, 4th order Runge-Kutta, and Gear orders 1 through 4.



#ifndef INTEGRATORS_H
#define INTEGRATORS_H

#include <integrator_def.h>

#include <deque>
#include <stdexcept>


// Generic integrator constructor

// Arguments:
//        x0        initial state vector
//        t0        initial time
//        h_init    step size
//        dxdt_calc function which calculates
//                     the rhs's of the DEs

template <class C, class T>
    Integrator<C,T>::Integrator<C,T> (const C &x0, const T t0,
        const T h_init, 
        void (*dxdt_calc) (const C&, const T, C &))
        :statevector (x0)
{
    h = h_init;
    t = t0;
    return_derivs = dxdt_calc;
}

// Function to retrieve the current state vector and time.
// Arguments:    
//          curr_x  Vector in which to store the current
//                    state variables
//          curr_t  Reference in which to store the current
//                    time
template <class C, class T>
    void Integrator<C,T>::get_state(C &curr_x, T &curr_t) const
{
    curr_t = t;
    curr_x = statevector;
}



// This function takes a single Euler step of size h.
template <class C, class T, class B> void
Euler<C,T,B>::step(void)
{
    // Calculate the derivatives.
    return_derivs(statevector,t,dxdt);

    // Update the variables and t.
    statevector += B(h)*dxdt;
    t += h;
}



// The step function for the 4th order Runge-Kutta integrator.
template <class C, class T, class B> void
RK4<C,T,B>::step (void)
{
    return_derivs (statevector, t, k1);
    return_derivs (statevector + B(h/2.0)*k1, t+ h/2.0, k2);
    return_derivs (statevector + B(h/2.0)*k2, t+ h/2.0, k3);
    return_derivs (statevector + B(h)*k3, t + h, k4);

//  statevector += h/6*(k1 + 2*k2 + 2*k3 + k4)
    k2 += k3;
    k2 *= B(2.0);
    k1 += k2;
    k1 += k4;
    k1 *= B(h/6.0);
    statevector += k1;

    t += h;
}


// Constructor for the Gear integrator
// Arguments: Gear_order     order of the method
//            x0             initial state vector
//            t0             initial time
//            h_init         initial step size
//            maxiterations  maximum number of iterations
//                           required to reach self-consistency
//            dxdt_calc      function which returns the
//                           RHS of the ODE
//            almost         true if the two vectors
//                           presented are almost equal
template <class C, class T, class B>
Gear<C,T,B>::Gear(const unsigned int Gear_order, const C &x0,
                  const T t0, const T h_init, 
                  const unsigned int maxiterations,
                  void (*dxdt_calc)(const C&, const T, C&),
                  bool (*almost)(const C&, const C&))
    :Integrator<C,T>(x0, t0, h_init, dxdt_calc),
     order(Gear_order), maxiter(maxiterations), xnext(x0),
     xtemp(x0), dxdt(x0)
{
    if (order > 4) throw std::runtime_error
             ("Bad order for the Gear integrator.");
    approx_equal = almost;
    curr_order = 1;
    calc_xnext[0] = &Gear<C,T,B>::Gear1xnext;
    calc_xnext[1] = &Gear<C,T,B>::Gear2xnext;
    calc_xnext[2] = &Gear<C,T,B>::Gear3xnext;
    calc_xnext[3] = &Gear<C,T,B>::Gear4xnext;
}


// Gear integrator
// Throws runtime_error if the iterations don't converge.
template <class C, class T, class B> void
Gear<C,T,B>::step(void)
{
    // Iterate until xnext is self-consistent.
    xnext = statevector;
    unsigned int iter = 0;
    do
    {
        xtemp = xnext;
        return_derivs(xnext,t+h,dxdt);
        (this->*(calc_xnext[curr_order-1]))();
        iter++;
        if (iter > maxiter)
            throw std::runtime_error("Too many iterations in Gear method.");
    } while (!approx_equal(xtemp,xnext));

    // Store the previous statevector.
    past_state.push_back(statevector);

    // If we're just using this integrator to bootstrap a
    // higher-order method, arrange to transfer control to 
    // the next order method.
    if (order > curr_order) curr_order++;
    else
    {
        // Get rid of the (now useless) oldest value of the past
        // state.
        past_state.pop_front();
    }

    // Update statevector and t.
    statevector = xnext;
    t += h;
}

// Reset the Gear integrator.
template <class C, class T, class B> void
Gear<C,T,B>::set_step_size(T new_h)
{
    // Empty the queue of past states.
    past_state.clear();

    // Reset to use the 1st order method.
    curr_order = 1;

    // Reset h.
    h = new_h;
}

// Gear formulas
// First order: xnext = statevector + h*dxdt
template <class C, class T, class B> void Gear<C,T,B>::Gear1xnext(void)
{
    xnext = dxdt;       xnext *= B(h);  xnext += statevector;
}

// Second order: xnext = (4*statevector - past_state[0] + 2*h*dxdt)/3
template <class C, class T, class B> void Gear<C,T,B>::Gear2xnext(void)
{
    xnext = dxdt;       xnext *= B(2.0*h);      xnext += B(4.0)*statevector;
    xnext -= past_state[0];     xnext /= B(3.0);
}

// Third order: 
// xnext = (18*statevector - 9*past_state[1] + 2*past_state[0] + 6*h*dxdt)/11
template <class C, class T, class B> void Gear<C,T,B>::Gear3xnext(void)
{
    xnext = statevector;        xnext *= B(2.0);        xnext -= past_state[1];
    xnext *= B(9.0);    xnext += B(2.0)*past_state[0];
    xnext += B(6.0*h)*dxdt;     xnext /= B(11.0);
}

// Fourth order:
// xnext = (48*statevector - 36*past_state[2] + 16*past_state[1]
//          - 3*past_state[0] + 12*h*dxdt)/25
template <class C, class T, class B> void Gear<C,T,B>::Gear4xnext(void)
{
    xnext = statevector;        xnext *= B(3.0);        xnext += past_state[1];
    xnext *= B(16.0);   xnext -= B(36.0)*past_state[2];
    xnext -= B(3.0)*past_state[0];      xnext += B(12.0*h)*dxdt;
    xnext /= B(25.0);
}

#endif    // Multiple inclusion guard
