/* From Meyers and Plum. "The New C: Low-Level I/O for Embedded Systems,"
 * C/C++ Users Journal, November 2003.
 *
 * This file defines the "functions" that operate on I/O Register designators
 * and I/O group designators specified in:
 * ISO/IEC TR 18037, C Extensions to Support Embedded Processors.
 *
 * Only the functions used by our example program are implemented.
 *
 * Note: All names beginning with _ are private implementation details
 * only to be used by iohw.h, iohw.c, iolpt.h, and iolpt.c.
 *
 */

#include "iohw.h"
#include "iolpt.h"

#include <dos.h>	/* DMC port I/O extensions */
#include <stdio.h>
#include <stdlib.h>

extern void iogroup_acquire(_IOgroup group)
{
    /* The LPT1 registers are dynamic designators since their port
     * numbers vary from machine to machine.  The BIOS tells us where
     * the port is located.
     *
     * In a more complete example, this would be the place to set the
     * I/O privilege mask to also grant access to port when the CPU
     * is in protected instead of real mode.  */
    if (group == LPT1_GROUP) {
	//unsigned short *bios_table = (unsigned short *) 0x400008;
	unsigned short *bios_table = (unsigned short *) _MK_FP(0x40, 0x08);
	if (bios_table[0] == 0) {
	    printf("No printer port\n");
	    exit(EXIT_FAILURE);
	}
	printf("Found printer port at %X\n", bios_table[0]);
	lpt1_data.location.port = bios_table[0];
	lpt1_status.location.port = bios_table[0] + 1;
	lpt1_control.location.port = bios_table[0] + 2;
	/* Port at 0x3BC never has a ECP control register.  Our standard
	 * check for a true ECP control register will cause us not to use
	 * it if it appears to be a copy of the SPP control register. */
	if (bios_table[0] == 0x3BC)
	    lpt1_ecp_control.location.port = lpt1_control.location.port;
	else 
	    lpt1_ecp_control.location.port = bios_table[0] + 0x402;
    }
}

extern void iogroup_release(_IOgroup group)
{
    /* Nothing to do.  */
}

extern unsigned int iord(_IOreg reg)
{
    switch (reg.type) {
      case _Port8:
	return _inp(reg.location.port);

      case _Port16:
	return _inpw(reg.location.port);

      case _Mem8:
	return *reg.location.ptr8;

      case _Mem16:
	return *reg.location.ptr16;

      default:
	abort();
    }
}

extern void iowr(_IOreg reg, unsigned int value)
{
    switch (reg.type) {
      case _Port8:
	_outp(reg.location.port, value);
	break;

      case _Port16:
	_outpw(reg.location.port, value);
	break;

      case _Mem8:
	*reg.location.ptr8 = value;
	break;

      case _Mem16:
	*reg.location.ptr16 = value;
	break;

      default:
	abort();
    }
}

extern void ioand(_IOreg reg, unsigned int value)
{
    /* Since we have no access to this operation in one instruction,
     * do a separate read and write.  */
    iowr(reg, iord(reg) & value);
}

extern void ioor(_IOreg reg, unsigned int value)
{
    /* Since we have no access to this operation in one instruction,
     * do a separate read and write.  */
    iowr(reg, iord(reg) | value);
}
