/* From Meyers and Plum. "The New C: Low-Level I/O for Embedded Systems,"
 * C/C++ Users Journal, November 2003.
 *
 * This example uses a mock up of the <iohw.h> and low-level I/O
 * extensions specified in
 *     ISO/IEC TR 18037, C Extensions to Support Embedded Processors.
 * to write a string to the first parallel port.  See readme.txt for
 * details.
 *
 */

#include "iohw.h"
#include "iolpt.h"

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>	/* Turbo C delay() */

void init_lpt1(void);
void printstr(char *sting);
void printchr(char c);

int main()
{
    /* Set up the LPT1 device registers for our use */
    iogroup_acquire(LPT1_GROUP);

    /* Initialize the parallel port and printer */
    init_lpt1();

    printstr("Hello world\r\n\f");

    iogroup_release(LPT1_GROUP);
    return EXIT_SUCCESS;
}

void init_lpt1(void)
{
    /* Disable interrupts, bidirectional mode, etc. */
    iowr(lpt1_control, 0);

    /* One recommended way to verify that the ECP control register exists
     * is to check if the resting port has an empty FIFO and that the full
     * and empty bits differ from the bits in the same position in the SPP
     * control register (we just set those to 0).  */
    if ((iord(lpt1_ecp_control) & (FIFO_FULL | FIFO_EMPTY)) == FIFO_EMPTY) {
	/* Set the port to Standard Parallel Port mode */
	iowr(lpt1_ecp_control, 0);
	printf("ECP control register set to SPP mode\n");
    }

    /* Initialize the printer and select it for I/O.  Leave these bits on
     * while we do I/O to the printer.  */
    iowr(lpt1_control, RESET|SELECT);
}

void printstr(char *string)
{
    char c;
    while ((c = *string++) != '\0')
	printchr(c);
}

void printchr(char c)
{
    long retry = 0;

    /* Set the data pins on the parallel port to the char */
    iowr(lpt1_data, c);

    /* If the printer is busy, it will miss the STROBE */
    while (!(iord(lpt1_status) & NOTBUSY)) {
	/* retry for up to 100 seconds */
	if (++retry > 100000) {
	    printf("Printer busy.  Exiting.\n");
	    exit(EXIT_FAILURE);
	}
	delay(1);
    }

    /* Turn on the STROBE pin on the parallel port.  This causes the
     * printer to read the data pins on the port. */
    ioor(lpt1_control, STROBE);

    /* We must hold STROBE for a time in microseconds, but we'll hold it
     * for a millisecond since that's our smallest delay. */
    delay(1);

    /* Turn off STROBE, thus getting ready to write another char. */
    ioand(lpt1_control, ~STROBE);
}
